import unittest
from unittest.mock import MagicMock, patch
import numpy as np

from carbontracker import exceptions
from carbontracker.components.gpu import nvidia
from carbontracker.components.component import (
    Component,
    create_components,
    error_by_name,
)


class TestComponent(unittest.TestCase):
    @patch("carbontracker.components.component.component_names", return_value=["gpu"])
    @patch(
        "carbontracker.components.component.error_by_name",
        return_value=exceptions.GPUError("No GPU(s) available."),
    )
    @patch(
        "carbontracker.components.component.handlers_by_name",
        return_value=[MagicMock(spec=nvidia.NvidiaGPU)],
    )
    def test_init_valid_component(
        self, mock_handlers_by_name, mock_error_by_name, mock_component_names
    ):
        component = Component(name="gpu", pids=[], devices_by_pid=False)
        self.assertEqual(component.name, "gpu")
        self.assertEqual(component._handler, mock_handlers_by_name()[0]())

    def test_init_invalid_component(self):
        with self.assertRaises(exceptions.ComponentNameError):
            Component(name="unknown", pids=[], devices_by_pid=False)

    def test_devices(self):
        handler_mock = MagicMock(devices=MagicMock(return_value=["Test GPU"]))
        component = Component(name="gpu", pids=[], devices_by_pid=False)
        component._handler = handler_mock
        self.assertEqual(component.devices(), ["Test GPU"])

    def test_available_true(self):
        handler_mock = MagicMock(available=MagicMock(return_value=True))
        component = Component(name="gpu", pids=[], devices_by_pid=False)
        component._handler = handler_mock
        self.assertTrue(component.available())

    @patch(
        "carbontracker.components.gpu.nvidia.NvidiaGPU.available", return_value=False
    )
    @patch(
        "carbontracker.components.apple_silicon.powermetrics.AppleSiliconGPU.available",
        return_value=False,
    )
    def test_available_false(self, mock_apple_gpu_available, mock_nvidia_gpu_available):
        component = Component(name="gpu", pids=[], devices_by_pid=False)
        self.assertFalse(component.available())

    def test_collect_power_usage_no_measurement(self):
        handler_mock = MagicMock(
            power_usage=MagicMock(side_effect=exceptions.IntelRaplPermissionError)
        )
        component = Component(name="cpu", pids=[], devices_by_pid=False)
        component._handler = handler_mock
        component.collect_power_usage(epoch=1)
        self.assertEqual(component.power_usages, [[], [0]])

    def test_collect_power_usage_with_measurement(self):
        handler_mock = MagicMock(power_usage=MagicMock(return_value=[1000]))
        component = Component(name="cpu", pids=[], devices_by_pid=False)
        component._handler = handler_mock
        component.collect_power_usage(epoch=1)
        self.assertEqual(component.power_usages, [[1000]])

    def test_collect_power_usage_with_measurement_but_no_epoch(self):
        power_collector = Component(name="cpu", pids=[], devices_by_pid=False)
        power_collector._handler = MagicMock(power_usage=MagicMock(return_value=[1000]))
        power_collector.collect_power_usage(epoch=0)
        assert len(power_collector.power_usages) == 0

    def test_collect_power_usage_with_previous_measurement(self):
        power_collector = Component(name="cpu", pids=[], devices_by_pid=False)
        power_collector._handler = MagicMock(power_usage=MagicMock(return_value=[1000]))
        power_collector.collect_power_usage(epoch=1)
        power_collector.collect_power_usage(epoch=3)
        assert len(power_collector.power_usages) == 3

    def test_collect_power_usage_GPUPowerUsageRetrievalError(self):
        handler_mock = MagicMock(
            power_usage=MagicMock(side_effect=exceptions.GPUPowerUsageRetrievalError)
        )
        component = Component(name="gpu", pids=[], devices_by_pid=False)
        component._handler = handler_mock
        component.collect_power_usage(epoch=1)
        self.assertEqual(component.power_usages, [[], [0]])

    def test_energy_usage(self):
        component = Component(name="cpu", pids=[], devices_by_pid=False)
        component.power_usages = [[1000], [2000], [3000]]
        epoch_times = [1, 2, 3]
        energy_usages = component.energy_usage(epoch_times)
        self.assertEqual(
            energy_usages, [0.0002777777777777778, 0.0011111111111111111, 0.0025]
        )
        self.assertTrue(np.all(np.array(energy_usages) > 0))

    def test_energy_usage_no_measurements(self):
        component = Component(name="cpu", pids=[], devices_by_pid=False)
        component.power_usages = [[]]
        epoch_times = [1]
        energy_usages = component.energy_usage(epoch_times)
        self.assertEqual(energy_usages, [0])

    def test_energy_usage_with_power_from_later_epoch(self):
        component = Component(name="cpu", pids=[], devices_by_pid=False)
        component.power_usages = [[1000], [2000], [3000]]
        epoch_times = [1, 2, 3, 4]
        energy_usages = component.energy_usage(epoch_times)
        self.assertEqual(
            energy_usages,
            [0.0002777777777777778, 0.0011111111111111111, 0.0025, 0.0025],
        )

    def test_energy_usage_no_power(self):
        component = Component(name="cpu", pids=[], devices_by_pid=False)
        component.power_usages = [[], [], [], [], []]
        epoch_times = [1, 2, 3, 4, 5]
        energy_usages = component.energy_usage(epoch_times)
        expected_energy_usages = [0, 0, 0, 0, 0]
        assert np.allclose(
            energy_usages, expected_energy_usages, atol=1e-8
        ), f"Expected {expected_energy_usages}, but got {energy_usages}"

    def test_init(self):
        handler_mock = MagicMock()
        component = Component(name="gpu", pids=[], devices_by_pid=False)
        component._handler = handler_mock
        component.init()
        handler_mock.init.assert_called_once()

        self.assertEqual(component.name, "gpu")
        self.assertEqual(component._handler, handler_mock)
        self.assertEqual(component.power_usages, [])
        self.assertEqual(component.cur_epoch, -1)

    def test_shutdown(self):
        handler_mock = MagicMock()
        component = Component(name="gpu", pids=[], devices_by_pid=False)
        component._handler = handler_mock
        component.shutdown()
        handler_mock.shutdown.assert_called_once()

    def test_create_components(self):
        gpu = create_components("gpu", pids=[], devices_by_pid=False)
        cpu = create_components("cpu", pids=[], devices_by_pid=False)
        all_components = create_components("all", pids=[], devices_by_pid=False)
        self.assertEqual(len(gpu), 1)
        self.assertEqual(len(cpu), 1)
        self.assertEqual(len(all_components), 2)

    def test_error_by_name(self):
        self.assertEqual(
            str(error_by_name("gpu")), str(exceptions.GPUError("No GPU(s) available."))
        )
        self.assertEqual(
            str(error_by_name("cpu")), str(exceptions.CPUError("No CPU(s) available."))
        )

    def test_handler_property_with_handler_set(self):
        component = Component(name="gpu", pids=[], devices_by_pid=False)
        component._handler = "test"
        self.assertEqual(component.handler, "test")

    def test_handler_property_without_handler(self):
        component = Component(name="gpu", pids=[], devices_by_pid=False)
        component._handler = None
        with self.assertRaises(exceptions.GPUError):
            component.handler()


if __name__ == "__main__":
    unittest.main()



import argparse
import subprocess
from carbontracker.tracker import CarbonTracker
from carbontracker import parser
import ast


def parse_logs(log_dir):
    parser.print_aggregate(log_dir=log_dir)


def main():
    """
    The **carbontracker** CLI allows the user to track the energy consumption and carbon intensity of any program.
    [Make sure that you have relevant permissions before running this.](/#permissions)

    Args:
        --log_dir (path, optional): Log directory. Defaults to `./logs`.
        --api_keys (str, optional): API keys in a dictionary-like format, e.g. `\'{"electricitymaps": "YOUR_KEY"}\'`
        --parse (path, optional): Directory containing the log files to parse.

    Example:
        Tracking the carbon intensity of `script.py`.

            $ carbontracker python script.py

        With example options

            $ carbontracker --log_dir='./logs' --api_keys='{"electricitymaps": "API_KEY_EXAMPLE"}' python script.py

        Parsing logs:

            $ carbontracker --parse ./internal_logs
    """

    # Create a parser for the known arguments
    cli_parser = argparse.ArgumentParser(description="CarbonTracker CLI", add_help=True)
    cli_parser.add_argument("--log_dir", type=str, default="./logs", help="Log directory")
    cli_parser.add_argument(
        "--api_keys",
        type=str,
        help="API keys in a dictionary-like format, e.g., "
             '\'{"electricitymaps": "YOUR_KEY"}\'',
        default=None,
    )
    cli_parser.add_argument("--parse", type=str, help="Directory containing the log files to parse.")

    # Parse known arguments only
    known_args, remaining_args = cli_parser.parse_known_args()

    # Check if the --parse argument is provided
    if known_args.parse:
        parse_logs(known_args.parse)
        return

    # Parse the API keys string into a dictionary
    api_keys = ast.literal_eval(known_args.api_keys) if known_args.api_keys else None

    tracker = CarbonTracker(
        epochs=1, log_dir=known_args.log_dir, epochs_before_pred=0, api_keys=api_keys
    )
    tracker.epoch_start()

    # The remaining_args are considered as the command to execute
    if remaining_args:
        try:
            # Execute the command
            subprocess.run(remaining_args, check=True)
        except subprocess.CalledProcessError as e:
            print(f"Error executing command: {' '.join(remaining_args)}")
            print(f"Subprocess error: {e}")

    tracker.epoch_end()
    tracker.stop()


if __name__ == "__main__":
    main()


import os
import re
import time

from carbontracker import exceptions
from carbontracker.components.handler import Handler
from typing import List, Union

# RAPL Literature:
# https://www.researchgate.net/publication/322308215_RAPL_in_Action_Experiences_in_Using_RAPL_for_Power_Measurements

RAPL_DIR = "/sys/class/powercap/"
CPU = 0
DRAM = 2
MEASURE_DELAY = 1


class IntelCPU(Handler):
    def __init__(self, pids: List, devices_by_pid: bool):
        super().__init__(pids, devices_by_pid)
        self._handler = None

    def devices(self) -> List[str]:
        """Returns the name of all RAPL Domains"""
        return self._devices

    def available(self) -> bool:
        return os.path.exists(RAPL_DIR) and bool(os.listdir(RAPL_DIR))

    def power_usage(self) -> List[float]:
        before_measures = self._get_measurements()
        time.sleep(MEASURE_DELAY)
        after_measures = self._get_measurements()
        # Ensure all power measurements >= 0 and retry up to 3 times.
        attempts = 3
        while attempts > 0:
            attempts -= 1
            power_usages = [
                self._compute_power(before, after)
                for before, after in zip(before_measures, after_measures)
            ]
            if all(power >= 0 for power in power_usages):
                return power_usages
        default = [0.0 for device in range(len(self._devices))]
        return default

    def _compute_power(self, before: int, after: int) -> float:
        """Compute avg. power usage from two samples in microjoules."""
        joules = (after - before) / 1000000
        watt = joules / MEASURE_DELAY
        return watt

    def _read_energy(self, path: str) -> int:
        with open(os.path.join(path, "energy_uj"), "r") as f:
            return int(f.read())

    def _get_measurements(self):
        measurements = []
        for package in self._rapl_devices:
            try:
                power_usage = self._read_energy(os.path.join(RAPL_DIR, package))
                measurements.append(power_usage)
            # If there is no sudo access, we cannot read the energy_uj file.
            # Permission denied error is raised.
            except PermissionError:
                raise exceptions.IntelRaplPermissionError()

            except FileNotFoundError:
                # check cpu/gpu/dram
                parts = [
                    f
                    for f in os.listdir(os.path.join(RAPL_DIR, package))
                    if re.match(self.parts_pattern, f)
                ]
                total_power_usage = 0
                for part in parts:
                    total_power_usage += self._read_energy(
                        os.path.join(RAPL_DIR, package, part)
                    )

                measurements.append(total_power_usage)

        return measurements

    def _convert_rapl_name(self, name, pattern) -> Union[None, str]:
        if re.match(pattern, name):
            return "cpu:" + name[-1]

    def init(self):
        # Get amount of intel-rapl folders
        packages = list(filter(lambda x: ":" in x, os.listdir(RAPL_DIR)))
        self.device_count = len(packages)
        self._devices: List[str] = []
        self._rapl_devices: List[str] = []
        self.parts_pattern = re.compile(r"intel-rapl:(\d):(\d)")
        devices_pattern = re.compile("intel-rapl:.")

        for package in packages:
            if re.fullmatch(devices_pattern, package):
                with open(os.path.join(RAPL_DIR, package, "name"), "r") as f:
                    name = f.read().strip()
                if name != "psys":
                    self._rapl_devices.append(package)
                    rapl_name = self._convert_rapl_name(package, devices_pattern)
                    if rapl_name is not None:
                        self._devices.append(rapl_name)

    def shutdown(self):
        pass



import unittest
from carbontracker.components.handler import Handler

class TestHandler(unittest.TestCase):
    def setUp(self):
        # Create a Handler instance
        self.handler = Handler(pids=[], devices_by_pid={})

    def test_devices_raises_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.handler.devices()

    def test_available_raises_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.handler.available()

    def test_power_usage_raises_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.handler.power_usage()

    def test_init_raises_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.handler.init()

    def test_shutdown_raises_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.handler.shutdown()


if __name__ == '__main__':
    unittest.main()



import unittest
from carbontracker import exceptions

class TestExceptions(unittest.TestCase):
    def test_no_components_available_error(self):
        with self.assertRaises(exceptions.NoComponentsAvailableError):
            raise exceptions.NoComponentsAvailableError

    def test_unit_error(self):
        with self.assertRaises(exceptions.UnitError):
            raise exceptions.UnitError("Expected", "Received", "Message")

    def test_intel_rapl_permission_error(self):
        with self.assertRaises(exceptions.IntelRaplPermissionError):
            raise exceptions.IntelRaplPermissionError

    def test_gpu_power_usage_retrieval_error(self):
        with self.assertRaises(exceptions.GPUPowerUsageRetrievalError):
            raise exceptions.GPUPowerUsageRetrievalError

    def test_carbon_intensity_fetcher_error(self):
        with self.assertRaises(exceptions.CarbonIntensityFetcherError):
            raise exceptions.CarbonIntensityFetcherError

    def test_ip_location_error(self):
        with self.assertRaises(exceptions.IPLocationError):
            raise exceptions.IPLocationError

    def test_gpu_error(self):
        with self.assertRaises(exceptions.GPUError):
            raise exceptions.GPUError

    def test_cpu_error(self):
        with self.assertRaises(exceptions.CPUError):
            raise exceptions.CPUError

    def test_component_name_error(self):
        with self.assertRaises(exceptions.ComponentNameError):
            raise exceptions.ComponentNameError

    def test_fetcher_name_error(self):
        with self.assertRaises(exceptions.FetcherNameError):
            raise exceptions.FetcherNameError

    def test_mismatched_log_files_error(self):
        with self.assertRaises(exceptions.MismatchedLogFilesError):
            raise exceptions.MismatchedLogFilesError



from abc import ABCMeta, abstractmethod

""" 
    Information about the geocoder object g_location available
    here: https://geocoder.readthedocs.io
"""


class IntensityFetcher:
    __metaclass__ = ABCMeta

    @abstractmethod
    def suitable(self, g_location):
        """Returns True if it can be used based on geocoder location."""
        raise NotImplementedError

    @abstractmethod
    def carbon_intensity(self, g_location, time_dur=None):
        """
        Returns the carbon intensity by location and duration (s).
        If the API supports predicted intensities time_dur can be used.
        """
        raise NotImplementedError



class Location:
    # geocoder has no type hints, so this class represents the "location" object
    def __init__(self, ok: bool, address: str, country: str):
        self.ok = ok
        self.address = address
        self.country = country



import sys
import unittest
from unittest.mock import patch, MagicMock
import pynvml
from carbontracker import exceptions
from carbontracker.components.gpu.nvidia import NvidiaGPU


class PynvmlStub:
    @staticmethod
    def nvmlInit():
        pass

    @staticmethod
    def nvmlShutdown():
        # NvidiaGPU._handles = None
        pass

    @staticmethod
    def nvmlDeviceGetHandleByIndex(index):
        return index

    @staticmethod
    def nvmlDeviceGetCount():
        return 1

    @staticmethod
    def nvmlDeviceGetPowerUsage(handle):
        return 1000  # Returns power usage in mW

    @staticmethod
    def nvmlDeviceGetName(handle):
        if sys.version_info < (3, 10):
            return b"GPU"
        else:
            return "GPU"

    @staticmethod
    def nvmlDeviceGetComputeRunningProcesses(handle):
        mock_process = MagicMock()
        mock_process.pid = 1234
        return [mock_process]

    @staticmethod
    def nvmlDeviceGetGraphicsRunningProcesses(handle):
        mock_process = MagicMock()
        mock_process.pid = 1234
        return [mock_process]


class TestNvidiaGPU(unittest.TestCase):
    @patch("carbontracker.components.gpu.nvidia.pynvml", new=PynvmlStub)
    def test_devices(self):
        gpu = NvidiaGPU(pids=[], devices_by_pid=False)
        gpu._handles = [0]
        self.assertEqual(gpu.devices(), ["GPU"])

    @patch("carbontracker.components.gpu.nvidia.pynvml", new=PynvmlStub)
    def test_available(self):
        gpu = NvidiaGPU(pids=[], devices_by_pid=False)
        self.assertTrue(gpu.available())

    @patch("carbontracker.components.gpu.nvidia.pynvml", new=PynvmlStub)
    def test_power_usage(self):
        gpu = NvidiaGPU(pids=[], devices_by_pid=False)
        gpu._handles = [0]
        self.assertEqual(gpu.power_usage(), [1])

    @patch("carbontracker.components.gpu.nvidia.pynvml", new=PynvmlStub)
    def test_init_shutdown(self):
        gpu = NvidiaGPU(pids=[], devices_by_pid=False)
        gpu.init()
        self.assertNotEqual(gpu._handles, [])
        gpu.shutdown()
        self.assertEqual(gpu._handles, [])

    @patch("carbontracker.components.gpu.nvidia.pynvml", new=PynvmlStub)
    def test_init(self):
        gpu = NvidiaGPU(pids=[1234], devices_by_pid=True)
        self.assertEqual(gpu.pids, [1234])
        self.assertEqual(gpu.devices_by_pid, True)
        self.assertEqual(gpu._handles, [])

    @patch("carbontracker.components.gpu.nvidia.pynvml", new=PynvmlStub)
    def test_get_handles(self):
        gpu = NvidiaGPU(pids=[], devices_by_pid=False)
        gpu.init()
        self.assertEqual(gpu._handles, [0])
        gpu.shutdown()

    @patch("carbontracker.components.gpu.nvidia.pynvml", new=PynvmlStub)
    @patch("carbontracker.components.gpu.nvidia.os.environ.get", return_value="0")
    def test_slurm_gpu_indices(self, mock_get):
        gpu = NvidiaGPU(pids=[], devices_by_pid=False)
        self.assertEqual(gpu._slurm_gpu_indices(), [0])

    @patch("carbontracker.components.gpu.nvidia.pynvml", new=PynvmlStub)
    def test_get_handles_by_pid(self):
        gpu = NvidiaGPU(pids=[1234], devices_by_pid=True)
        gpu.init()
        self.assertEqual(gpu._handles, [0])
        gpu.shutdown()

    @patch("sys.version_info", new=(3, 8))
    @patch("carbontracker.components.gpu.nvidia.pynvml", new=PynvmlStub)
    def test_devices_python_version_less_than_3_10(self):
        gpu = NvidiaGPU(pids=[], devices_by_pid=False)
        gpu._handles = [0]
        self.assertEqual(gpu.devices(), ["GPU"])

    @patch(
        "carbontracker.components.gpu.nvidia.pynvml.nvmlDeviceGetPowerUsage",
        side_effect=pynvml.NVMLError(pynvml.NVML_ERROR_UNKNOWN),
    )
    def test_power_usage_error_retrieving_power_usage(
        self, mock_nvmlDeviceGetPowerUsage
    ):
        gpu = NvidiaGPU(pids=[], devices_by_pid=False)
        gpu._handles = [0]
        with self.assertRaises(exceptions.GPUPowerUsageRetrievalError):
            gpu.power_usage()

    @patch(
        "carbontracker.components.gpu.nvidia.pynvml.nvmlDeviceGetComputeRunningProcesses",
        return_value=[],
    )
    @patch(
        "carbontracker.components.gpu.nvidia.pynvml.nvmlDeviceGetGraphicsRunningProcesses",
        return_value=[],
    )
    def test_get_handles_by_pid_no_gpus_running_processes(
        self,
        mock_nvmlDeviceGetComputeRunningProcesses,
        mock_nvmlDeviceGetGraphicsRunningProcesses,
    ):
        gpu = NvidiaGPU(pids=[1234], devices_by_pid=True)
        self.assertEqual(gpu._handles, [])

    @patch("carbontracker.components.gpu.nvidia.pynvml.nvmlInit")
    @patch(
        "carbontracker.components.gpu.nvidia.pynvml.nvmlDeviceGetCount", return_value=0
    )
    def test_available_no_gpus(self, mock_nvmlDeviceGetCount, mock_nvmlInit):
        gpu = NvidiaGPU(pids=[], devices_by_pid=False)
        self.assertFalse(gpu.available())

    @patch(
        "carbontracker.components.gpu.nvidia.pynvml.nvmlInit",
        side_effect=pynvml.NVMLError(pynvml.NVML_ERROR_UNKNOWN),
    )
    def test_available_nvml_error(self, mock_nvmlInit):
        gpu = NvidiaGPU(pids=[], devices_by_pid=False)
        self.assertFalse(gpu.available())


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
from carbontracker import predictor

class TestPredictor(unittest.TestCase):
    def test_predict_energy(self):
        total_epochs = 10
        epoch_energy_usages = np.array([10, 20, 30, 40, 50])
        expected_result = total_epochs * np.mean(epoch_energy_usages)
        result = predictor.predict_energy(total_epochs, epoch_energy_usages)
        self.assertEqual(result, expected_result)

    def test_predict_time(self):
        total_epochs = 10
        epoch_times = np.array([1, 2, 3, 4, 5])
        expected_result = total_epochs * np.mean(epoch_times)
        result = predictor.predict_time(total_epochs, epoch_times)
        self.assertEqual(result, expected_result)




import unittest
from carbontracker.emissions.conversion.co2eq import convert

class TestConversion(unittest.TestCase):
    def test_convert(self):
        expected = [(2.32558139535, 'km travelled by car')]
        actual = convert(250)
        self.assertAlmostEqual(expected[0][0], actual[0][0], places=5)
        self.assertEqual(expected[0][1], actual[0][1])  # compare unit names

        expected = [(4.6511627907, 'km travelled by car')]
        actual = convert(500)
        self.assertAlmostEqual(expected[0][0], actual[0][0], places=5)
        self.assertEqual(expected[0][1], actual[0][1])

if __name__ == '__main__':
    unittest.main()



import geocoder
import unittest
from unittest.mock import patch, MagicMock
import numpy as np
import pandas as pd
import sys

from carbontracker import constants
from carbontracker.emissions.intensity import intensity

from carbontracker.emissions.intensity.intensity import carbon_intensity


class TestIntensity(unittest.TestCase):
    @patch("geocoder.ip")
    def test_get_default_intensity_success(self, mock_geocoder_ip):
        mock_location = MagicMock()
        mock_location.ok = True
        mock_location.address = "Sample Address"
        mock_location.country = "US"
        mock_geocoder_ip.return_value = mock_location

        result = intensity.get_default_intensity()

        # importlib.resources.files was introduced in Python 3.9 and replaces deprecated pkg_resource.resources
        if sys.version_info < (3,9):
            import pkg_resources
            carbon_intensities_df = pd.read_csv(pkg_resources.resource_filename("carbontracker", "data/carbon-intensities.csv"))
        else:
            import importlib.resources
            ref = importlib.resources.files("carbontracker") / "data/carbon-intensities.csv"
            with importlib.resources.as_file(ref) as path:
                carbon_intensities_df = pd.read_csv(path)
        intensity_row = carbon_intensities_df[carbon_intensities_df["alpha-2"] == mock_location.country].iloc[0]
        expected_intensity = intensity_row["Carbon intensity of electricity (gCO2/kWh)"]

        self.assertEqual(result["carbon_intensity"], expected_intensity)
        self.assertIn("Defaulted to average carbon intensity", result["description"])

    @patch("geocoder.ip")
    def test_get_default_intensity_location_failure(self, mock_geocoder_ip):
        mock_geocoder_ip.return_value.ok = False

        result = intensity.get_default_intensity()

        self.assertEqual(result["carbon_intensity"], constants.WORLD_2019_CARBON_INTENSITY)
        self.assertIn("Defaulted to average carbon intensity", result["description"])

    @patch("geocoder.ip")
    @patch("pandas.read_csv")
    def test_get_default_intensity_data_file_failure(
            self, mock_read_csv, mock_geocoder_ip
    ):
        mock_location = MagicMock()
        mock_location.ok = True
        mock_location.address = "Sample Address"
        mock_location.country = "US"
        mock_geocoder_ip.return_value = mock_location

        mock_read_csv.side_effect = FileNotFoundError

        default_intensity = intensity.get_default_intensity()

        expected_description = (
            f"Live carbon intensity could not be fetched at detected location: {mock_location.address}. "
            f"Defaulted to average carbon intensity for world in 2019 of {constants.WORLD_2019_CARBON_INTENSITY:.2f} gCO2/kWh."
        )

        assert default_intensity["carbon_intensity"] == constants.WORLD_2019_CARBON_INTENSITY
        assert default_intensity["description"] == expected_description

    @patch("carbontracker.emissions.intensity.intensity.geocoder.ip")
    @patch("carbontracker.emissions.intensity.intensity.pd.read_csv")
    def test_CarbonIntensity_set_as_default(
            self, mock_read_csv, mock_geocoder_ip
    ):
        mock_location = MagicMock()
        mock_location.ok = True
        mock_location.address = "Sample Address"
        mock_location.country = "US"
        mock_geocoder_ip.return_value = mock_location

        mock_read_csv.side_effect = FileNotFoundError

        default_intensity = intensity.get_default_intensity()

        expected_description = (
            f"Live carbon intensity could not be fetched at detected location: {mock_location.address}. "
            f"Defaulted to average carbon intensity for world in 2019 of {constants.WORLD_2019_CARBON_INTENSITY:.2f} gCO2/kWh."
        )

        self.assertEqual(default_intensity["carbon_intensity"], constants.WORLD_2019_CARBON_INTENSITY)
        self.assertEqual(default_intensity["description"], expected_description)


    def test_CarbonIntensity_set_default_message(self):
        ci = intensity.CarbonIntensity(default=True)
        ci.set_default_message()

        self.assertIn("Defaulted to average carbon intensity", ci.message)

    @patch("geocoder.ip")
    def test_get_default_intensity_ip_location_failure(self, mock_geocoder_ip):
        mock_geocoder_ip.return_value.ok = False

        result = intensity.get_default_intensity()

        self.assertEqual(result["carbon_intensity"], constants.WORLD_2019_CARBON_INTENSITY)
        self.assertIn("Defaulted to average carbon intensity", result["description"])

    @patch("geocoder.ip")
    def test_carbon_intensity_location_failure(self, mock_geocoder_ip):
        mock_geocoder_ip.return_value.ok = False

        logger = MagicMock()

        with patch('carbontracker.emissions.intensity.intensity.default_intensity', intensity.get_default_intensity()) as C:
            result = intensity.carbon_intensity(logger)
            default_intensity = intensity.get_default_intensity()

            self.assertEqual(result.carbon_intensity, default_intensity["carbon_intensity"])
            self.assertEqual(result.address, "UNDETECTED")
            self.assertEqual(result.success, False)
            self.assertIn("Live carbon intensity could not be fetched at detected location", result.message)

    @patch("carbontracker.emissions.intensity.intensity.geocoder.ip")
    def test_set_carbon_intensity_message(self, mock_geocoder_ip):
        time_dur = 3600
        mock_location = MagicMock()
        # Assuming the actual function logic uses a specific fallback or detected location
        detected_address = "Aarhus, Capital Region, DK"  # The detected location that appears in the error
        fallback_address = "Generic Location, Country"  # The fallback or generic location
        mock_location.address = detected_address
        mock_location.country = 'DK'
        mock_geocoder_ip.ok = True
        mock_geocoder_ip.return_value = mock_location
        # Adjust the set_expected_message function to match the error details
        def set_expected_message(is_prediction, success, carbon_intensity):
            if is_prediction:
                if success:
                    message = f"Carbon intensity for the next 1:00:00 is predicted to be {carbon_intensity:.2f} gCO2/kWh at detected location: {fallback_address}."
                else:
                    message = f"Failed to predict carbon intensity for the next 1:00:00, fallback on average measured intensity at detected location: {fallback_address}."
            else:
                if success:
                    message = f"Current carbon intensity is {carbon_intensity:.2f} gCO2/kWh at detected location: {fallback_address}."
                else:
                    message = (f"Live carbon intensity could not be fetched at detected location: {detected_address}. "
                               f"Defaulted to average carbon intensity for DK in 2023 of 151.65 gCO2/kWh. "
                               f"at detected location: {fallback_address}.")
            return message
        # Test scenarios
        scenarios = [
            (True, True, 100.0),
            (True, False, None),
            (False, True, 50.0),
            (False, False, None)  # The scenario corresponding to the failure message
        ]

        with patch('carbontracker.emissions.intensity.intensity.default_intensity', intensity.get_default_intensity()) as C:
            ci = intensity.CarbonIntensity()
            ci.address = fallback_address  # Set to the fallback or generic address for the test case
            for is_prediction, success, carbon_intensity in scenarios:
                ci.is_prediction = is_prediction
                ci.success = success
                ci.carbon_intensity = carbon_intensity if carbon_intensity is not None else 0.0
                intensity.set_carbon_intensity_message(ci, time_dur)
                expected_message = set_expected_message(is_prediction, success, carbon_intensity)
                self.assertEqual(ci.message, expected_message)

    @patch("geocoder.ip")
    @patch("carbontracker.emissions.intensity.fetchers.electricitymaps.ElectricityMap.suitable")
    def test_carbon_intensity_address_assignment(self, mock_electricity_map_suitable, mock_geocoder_ip):
        mock_location = MagicMock()
        mock_location.ok = True
        mock_location.address = None
        mock_geocoder_ip.return_value = mock_location

        mock_electricity_map_suitable.return_value = False

        logger = MagicMock()
        result = carbon_intensity(logger)

        self.assertIsNone(result.address)
        mock_electricity_map_suitable.assert_called_once_with(mock_location)

    @patch("geocoder.ip")
    @patch("carbontracker.emissions.intensity.fetchers.electricitymaps.ElectricityMap.carbon_intensity")
    def test_carbon_intensity_failure(self, mock_electricity_map_carbon_intensity, mock_geocoder_ip):
        mock_location = MagicMock()
        mock_location.ok = True
        mock_geocoder_ip.return_value = mock_location

        mock_electricity_map_carbon_intensity.side_effect = Exception("Test Exception")

        logger = MagicMock()

        result = carbon_intensity(logger)

        self.assertFalse(result.success)
        self.assertIn("could not be fetched", result.message)

    @patch("carbontracker.emissions.intensity.fetchers.carbonintensitygb.CarbonIntensityGB")
    @patch("carbontracker.emissions.intensity.intensity.geocoder.ip")
    def test_carbon_intensity_exception_carbonintensitygb(self, mock_geocoder, mock_carbonintensitygb):
        mock_geocoder.return_value.address = "Sample Address"
        mock_geocoder.return_value.ok = True
        mock_carbonintensitygb.return_value.suitable.return_value = True

        mock_result = MagicMock()
        mock_result.carbon_intensity = 23.0
        mock_result.success = True

        mock_carbonintensitygb.return_value.carbon_intensity.return_value = mock_result

        logger = MagicMock()

        result = carbon_intensity(logger, fetchers=[mock_carbonintensitygb()])

        self.assertEqual(result.carbon_intensity, 23.0)
        self.assertTrue(result.success)

    @patch("carbontracker.emissions.intensity.fetchers.energidataservice.EnergiDataService")
    def test_carbon_intensity_energidataservice(self, mock_energidataservice):
        mock_energidataservice.return_value.suitable.return_value = True

        mock_result = MagicMock()
        mock_result.carbon_intensity = 23.0
        mock_result.success = True
        mock_energidataservice.return_value.carbon_intensity.return_value = mock_result

        logger = MagicMock()
        result = carbon_intensity(logger, fetchers=[mock_energidataservice()])

        self.assertEqual(result.carbon_intensity, 23.0)
        self.assertTrue(result.success)

    @patch("carbontracker.emissions.intensity.intensity.geocoder.ip")
    @patch("carbontracker.emissions.intensity.fetchers.electricitymaps.ElectricityMap")
    def test_carbon_intensity_nan(self, mock_electricity_map, mock_geocoder):
        mock_location = MagicMock()
        mock_location.ok = True
        mock_location.address = "Sample Address"
        mock_geocoder.ip.return_value = mock_location

        mock_electricity_map.return_value.suitable.return_value = True

        mock_result = MagicMock()
        mock_result.carbon_intensity = np.nan
        mock_result.success = False

        mock_electricity_map.return_value.carbon_intensity.return_value = mock_result

        logger = MagicMock()

        result = carbon_intensity(logger)

        self.assertFalse(result.success)
        self.assertTrue(np.isnan(result.carbon_intensity))
        self.assertEqual(mock_location.address, "Sample Address")



import unittest
from unittest import skipIf
from carbontracker import loggerutil
from carbontracker.loggerutil import Logger, convert_to_timestring
import unittest.mock
from unittest.mock import MagicMock, patch
import tempfile
import os
import logging
from datetime import datetime
import time


class TestLoggerUtil(unittest.TestCase):
    def test_convert_to_timestring_positive_value(self):
        time_s = 3666
        self.assertEqual(convert_to_timestring(time_s), "1:01:06")

    def test_convert_to_timestring_zero_value(self):
        time_s = 0
        self.assertEqual(convert_to_timestring(time_s), "0:00:00")

    def test_convert_to_timestring_floating_value(self):
        time_s = 3666.4
        self.assertEqual(convert_to_timestring(time_s), "1:01:06")

    def test_convert_to_timestring_negative_value(self):
        time_s = -3666
        self.assertEqual(convert_to_timestring(time_s), "-1:01:06")

    def test_convert_to_timestring_rounding_seconds(self):
        time_s = 3659.5  # This will round to 3660, which is 1 hour and 60 seconds
        self.assertEqual(convert_to_timestring(time_s), "1:01:00")

    def test_convert_to_timestring_rounding_float_seconds(self):
        time_s = 3659.9955  # Very close to 3660, and should round off to it
        self.assertEqual(
            convert_to_timestring(time_s, add_milliseconds=True), "1:01:00.00"
        )

    @skipIf(os.environ.get("CI") == "true", "Skipped due to CI")
    def test_formatTime_with_datefmt(self):
        formatter = loggerutil.TrackerFormatter()
        record = MagicMock()
        record.created = time.mktime(
            datetime(2023, 3, 15, 14, 20, 0).timetuple()
        )  # This is a sample timestamp for "2023-03-15 14:20:00" at UTC time

        # Specify a custom date format
        datefmt = "%Y-%m-%d %H-%M-%S"
        formatted_time = formatter.formatTime(record, datefmt)

        self.assertEqual(formatted_time, "2023-03-15 14-20-00")

    @skipIf(os.environ.get("CI") == "true", "Skipped due to CI")
    def test_formatTime_without_datefmt(self):
        formatter = loggerutil.TrackerFormatter()
        record = MagicMock()
        record.created = time.mktime(datetime(2023, 3, 15, 14, 20, 0).timetuple())

        formatted_time = formatter.formatTime(record)

        self.assertEqual(formatted_time, "2023-03-15 14:20:00")

    def test_logger_with_log_prefix(self):
        log_prefix_original = "test_prefix"
        logger = loggerutil.Logger(log_prefix=log_prefix_original)

        # Check if the logger's name starts with the updated log_prefix
        self.assertTrue(logger.logger.name.startswith(f"{log_prefix_original}_"))

    def test_logger_without_log_prefix(self):
        logger = loggerutil.Logger(log_prefix="")

        # Check if the logger's name does not contain any underscores (indicating no prefix was added)
        self.assertFalse("_" in logger.logger.name)

    def test_VerboseFilter_with_verbose(self):
        verbose_filter = loggerutil.VerboseFilter(verbose=1)
        record = MagicMock()

        # The filter should return True since verbose is set to 1
        self.assertTrue(verbose_filter.filter(record))

    def test_VerboseFilter_without_verbose(self):
        verbose_filter = loggerutil.VerboseFilter(verbose=0)
        record = MagicMock()

        # The filter should return False since verbose is set to 0
        self.assertFalse(verbose_filter.filter(record))

    def test_logger_setup(self):
        logger = Logger()
        self.assertIsInstance(logger, Logger)
        self.assertEqual(
            logger.logger_output.level, logging.DEBUG, "Logging level is not DEBUG."
        )

    def test_info_logging(self):
        logger = Logger()
        with unittest.mock.patch.object(logger.logger, "info") as mock_info:
            msg = "Test Info Message"
            logger.info(msg)
            mock_info.assert_called_once_with(msg)

    def test_err_debug_logging(self):
        logger = Logger()
        with unittest.mock.patch.object(logger.logger_err, "debug") as mock_debug:
            msg = "Test Debug Error Message"
            logger.err_debug(msg)
            mock_debug.assert_called_once_with(msg)

    def test_err_info_logging(self):
        logger = Logger()
        with unittest.mock.patch.object(logger.logger_err, "info") as mock_info:
            msg = "Test Info Error Message"
            logger.err_info(msg)
            mock_info.assert_called_once_with(msg)

    def test_err_warn_logging(self):
        logger = Logger()
        with unittest.mock.patch.object(logger.logger_err, "warning") as mock_warn:
            msg = "Test Warn Error Message"
            logger.err_warn(msg)
            mock_warn.assert_called_once_with(msg)

    def test_err_critical_logging(self):
        logger = Logger()
        with unittest.mock.patch.object(logger.logger_err, "critical") as mock_critical:
            msg = "Test Critical Error Message"
            logger.err_critical(msg)
            mock_critical.assert_called_once_with(msg)

    def test_log_initial_info(self):
        logger = Logger()
        with unittest.mock.patch.object(logger.logger, "info") as mock_info:
            logger._log_initial_info()  # Call it again for testing purposes
            self.assertEqual(
                mock_info.call_count, 2
            )  # Called twice: one during initialization and one during our test

    def test_logger_with_log_dir(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            logger = loggerutil.Logger(log_dir=tmp_dir)

            logger.err_info("Trigger error to create the log file")

            self.assertTrue(os.path.exists(tmp_dir))

            files = os.listdir(tmp_dir)
            self.assertTrue(any(["carbontracker_output.log" in file for file in files]))
            self.assertTrue(any(["carbontracker.log" in file for file in files]))
            self.assertTrue(any(["carbontracker_err.log" in file for file in files]))

    @unittest.mock.patch("logging.Logger.info")
    def test_output(self, mock_info):
        logger = loggerutil.Logger()
        test_message = "Test Message"

        # Reset the mock to ignore the initial calls to `info` during Logger initialization
        mock_info.reset_mock()

        logger.output(test_message)

        mock_info.assert_called_once_with(f"CarbonTracker: {test_message}")

    def test_multiple_loggers(self):
        logger1 = loggerutil.Logger(logger_id="1")
        logger2 = loggerutil.Logger(logger_id="2")
        self.assertNotEqual(logger1.logger, logger2.logger)
        self.assertNotEqual(logger1.logger_output, logger2.logger_output)
        self.assertNotEqual(logger1.logger_err, logger2.logger_err)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch, mock_open
from carbontracker.components.cpu.intel import IntelCPU
from carbontracker.components.component import Component
from carbontracker import exceptions
import re

class TestIntelCPU(unittest.TestCase):
    @patch("os.path.exists")
    @patch("os.listdir")
    def test_available(self, mock_listdir, mock_exists):
        mock_exists.return_value = True
        mock_listdir.return_value = ["some_directory"]

        component = Component(name='cpu', pids=[], devices_by_pid={})
        self.assertTrue(component.available())

    @patch("os.path.exists")
    @patch("os.listdir")
    @patch("builtins.open", new_callable=mock_open, read_data="some_name")
    def test_devices(self, mock_file, mock_listdir, mock_exists):
        mock_exists.return_value = True
        mock_listdir.side_effect = [["intel-rapl:0", "intel-rapl:1"], ["name"], ["name"]]

        cpu = IntelCPU(pids=[], devices_by_pid={})
        cpu.init()

        self.assertEqual(cpu.devices(), ["cpu:0", "cpu:1"])

    @patch("os.path.exists")
    @patch("os.listdir")
    @patch.object(Component, "available", return_value=False)
    def test_available_false(self, mock_available, mock_listdir, mock_exists):
        mock_exists.return_value = False
        mock_listdir.return_value = []

        cpu = Component(name='cpu', pids=[], devices_by_pid={})
        self.assertFalse(cpu.available())

    @patch("time.sleep")
    @patch("carbontracker.components.cpu.intel.IntelCPU._get_measurements")
    def test_power_usage_positive(self, mock_get_measurements, mock_sleep):
        mock_get_measurements.side_effect = [[10, 20], [20, 30]]
        mock_sleep.return_value = None

        cpu = IntelCPU(pids=[], devices_by_pid={})
        power_usages = cpu.power_usage()

        self.assertEqual(power_usages, [0.00001, 0.00001])

    @patch("time.sleep")
    @patch("carbontracker.components.cpu.intel.IntelCPU._get_measurements")
    def test_power_usage_negative(self, mock_get_measurements, mock_sleep):
        mock_get_measurements.side_effect = [[30, 20], [20, 30]]
        mock_sleep.return_value = None

        cpu = IntelCPU(pids=[], devices_by_pid={})
        cpu._devices = ["cpu:0", "cpu:1"]
        power_usages = cpu.power_usage()

        self.assertEqual(power_usages, [0.00, 0.00])


    @patch("builtins.open", new_callable=mock_open, read_data="1000000")
    def test__read_energy(self, mock_file):
        cpu = IntelCPU(pids=[], devices_by_pid={})
        energy = cpu._read_energy("/some/path")
        self.assertEqual(energy, 1000000)

    @patch("os.path.exists")
    @patch("os.listdir")
    @patch("builtins.open", new_callable=mock_open)
    def test__get_measurements(self, mock_file, mock_listdir, mock_exists):
        mock_exists.return_value = True
        mock_listdir.return_value = ["intel-rapl:0", "intel-rapl:1"]
        mock_file.return_value.read.return_value = "1000000"

        cpu = IntelCPU(pids=[], devices_by_pid={})
        cpu.init()

        measurements = cpu._get_measurements()
        self.assertEqual(measurements, [1000000, 1000000])

    @patch("os.listdir")
    @patch("builtins.open", new_callable=mock_open, read_data="cpu")
    def test__convert_rapl_name(self, mock_file, mock_listdir):
        mock_listdir.return_value = ["intel-rapl:0", "intel-rapl:1"]

        cpu = IntelCPU(pids=[], devices_by_pid={})
        cpu.init()

        self.assertEqual(cpu._convert_rapl_name("intel-rapl:0", re.compile("intel-rapl:.")), "cpu:0")

    @patch("os.path.exists")
    @patch("os.listdir")
    @patch("builtins.open", new_callable=mock_open, read_data="cpu")
    def test_init(self, mock_file, mock_listdir, mock_exists):
        mock_exists.return_value = True
        mock_listdir.return_value = ["intel-rapl:0", "intel-rapl:1"]

        cpu = IntelCPU(pids=[], devices_by_pid={})
        cpu.init()

        self.assertEqual(cpu.devices(), ["cpu:0", "cpu:1"])

    @patch("os.path.join")
    @patch("os.listdir")
    @patch("carbontracker.components.cpu.intel.IntelCPU._read_energy")
    def test__get_measurements_permission_error(self, mock_read_energy, mock_listdir, mock_path_join):
        mock_path_join.return_value = "/some/path"
        mock_read_energy.side_effect = PermissionError()

        cpu = IntelCPU(pids=[], devices_by_pid={})
        cpu._rapl_devices = ["device1"]
        with self.assertRaises(exceptions.IntelRaplPermissionError):
            cpu._get_measurements()

    @patch("os.path.join")
    @patch("os.listdir")
    @patch("carbontracker.components.cpu.intel.IntelCPU._read_energy")
    def test__get_measurements_file_not_found(self, mock_read_energy, mock_listdir, mock_path_join):
        mock_path_join.return_value = "/some/path"
        mock_read_energy.side_effect = [FileNotFoundError(), 1000000, 1000000, 1000000]
        mock_listdir.return_value = ["intel-rapl:0", "intel-rapl:1"]

        cpu = IntelCPU(pids=[], devices_by_pid={})
        cpu._rapl_devices = ["intel-rapl:0", "intel-rapl:1"]
        cpu.parts_pattern = re.compile(r"intel-rapl:.")
        measurements = cpu._get_measurements()

        self.assertEqual(measurements, [2000000, 1000000])


    def test_shutdown(self):
        cpu = IntelCPU(pids=[], devices_by_pid={})
        # As the shutdown method is currently a pass, there's nothing to assert here.
        # But we still call it for the sake of completeness and future modifications.
        cpu.shutdown()

if __name__ == "__main__":
    unittest.main()



import io
from unittest import mock
import os
from carbontracker import exceptions
from pyfakefs import fake_filesystem_unittest

from carbontracker import parser
from carbontracker.parser import (
    extract_measurements,
    parse_logs,
    print_aggregate,
    get_stats,
    parse_equivalents,
)


class TestParser(fake_filesystem_unittest.TestCase):
    def setUp(self):
        self.setUpPyfakefs()

    @mock.patch("os.listdir")
    @mock.patch("os.path.isfile")
    @mock.patch("os.path.getsize")
    def test_get_all_logs(self, mock_getsize, mock_isfile, mock_listdir):
        log_dir = "/path/to/logs"

        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker_output.log"),
            contents="output_log1 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "32487_2024-06-26T141608Z_carbontracker_output.log"),
            contents="output_log2 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker.log"),
            contents="std_log1 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "32487_2024-06-26T141608Z_carbontracker.log"),
            contents="std_log2 content",
        )

        mock_listdir.return_value = [
            "10151_2024-03-26T105926Z_carbontracker_output.log",
            "32487_2024-06-26T141608Z_carbontracker_output.log",
            "10151_2024-03-26T105926Z_carbontracker.log",
            "32487_2024-06-26T141608Z_carbontracker.log",
        ]

        mock_isfile.side_effect = lambda path: path.endswith(".log")
        mock_getsize.return_value = 100

        output_logs, std_logs = parser.get_all_logs(log_dir)

        expected_output_logs = [
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker_output.log"),
            os.path.join(log_dir, "32487_2024-06-26T141608Z_carbontracker_output.log"),
        ]
        expected_std_logs = [
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker.log"),
            os.path.join(log_dir, "32487_2024-06-26T141608Z_carbontracker.log"),
        ]

        self.assertCountEqual(output_logs, expected_output_logs)
        self.assertCountEqual(std_logs, expected_std_logs)

    @mock.patch("os.listdir")
    def test_get_devices(self, mock_open):
        log_data = """2022-11-14 15:42:43 - CarbonTracker: The following components were found: GPU with device(s) NVIDIA GeForce RTX 3060. CPU with device(s) cpu:0.
    2022-11-14 15:43:37 - CarbonTracker: 
    Actual consumption for 1 epoch(s):
    	Time:	0:00:53
    	Energy:	0.003504 kWh
    	CO2eq:	0.357451 g
    	This is equivalent to:
    	0.002969 km travelled by car
    2022-11-14 15:43:37 - CarbonTracker: 
    Predicted consumption for 4 epoch(s):
    	Time:	0:03:33
    	Energy:	0.014018 kWh
    	CO2eq:	1.429803 g
    	This is equivalent to:
    	0.011875 km travelled by car
    2022-11-14 15:43:37 - CarbonTracker: Finished monitoring."""

        mock_open.return_value.read.return_value = log_data

        devices = parser.get_devices(log_data)

        expected_devices = {"gpu": ["NVIDIA GeForce RTX 3060"], "cpu": ["cpu:0"]}

        self.assertEqual(devices, expected_devices)

    def test_get_epoch_durations(self):
        std_log_data = "2022-11-14 15:44:48 - Epoch 1:\nDuration: 0:02:21.90\n2022-11-14 15:44:48 - Epoch 2:\nDuration: 0:01:30"

        epoch_durations = parser.get_epoch_durations(std_log_data)

        expected_epoch_durations = [141.9, 90.0]

        self.assertEqual(epoch_durations, expected_epoch_durations)

    def test_get_avg_power_usages(self):
        std_log_data = (
            "2022-11-14 15:44:48 - Average power usage (W) for gpu: [136.86084615]\n"
            "2022-11-14 15:44:48 - Average power usage (W) for cpu: [13.389104]"
        )

        avg_power_usages = parser.get_avg_power_usages(std_log_data)

        expected_avg_power_usages = {"gpu": [[136.86084615]], "cpu": [[13.389104]]}

        self.assertEqual(avg_power_usages, expected_avg_power_usages)

    @mock.patch("os.listdir")
    @mock.patch("os.path.isfile")
    @mock.patch("os.path.getmtime")
    def test_get_most_recent_logs(self, mock_getmtime, mock_isfile, mock_listdir):
        log_dir = "/path/to/logs"

        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker_output.log"),
            contents="output_log1 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "32487_2024-06-26T141608Z_carbontracker_output.log"),
            contents="output_log2 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker.log"),
            contents="std_log1 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "32487_2024-06-26T141608Z_carbontracker.log"),
            contents="std_log2 content",
        )

        mock_listdir.return_value = [
            "10151_2024-03-26T105926Z_carbontracker_output.log",
            "32487_2024-06-26T141608Z_carbontracker_output.log",
            "10151_2024-03-26T105926Z_carbontracker.log",
            "32487_2024-06-26T141608Z_carbontracker.log",
        ]

        mock_isfile.side_effect = lambda path: path.endswith(".log")
        mock_getmtime.side_effect = [
            100,
            200,
            300,
            400,
        ]  # Mock the modification timestamps

        std_log, output_log = parser.get_most_recent_logs(log_dir)

        expected_std_log = os.path.join(
            log_dir, "32487_2024-06-26T141608Z_carbontracker.log"
        )
        expected_output_log = os.path.join(
            log_dir, "32487_2024-06-26T141608Z_carbontracker_output.log"
        )

        self.assertEqual(std_log, expected_std_log)
        self.assertEqual(output_log, expected_output_log)

    def test_get_time(self):
        time_str = "0:02:22"

        duration = parser.get_time(time_str)

        expected_duration = 142.0

        self.assertEqual(duration, expected_duration)

    def test_get_early_stop(self):
        std_log_data = "2022-11-14 15:44:48 - CarbonTracker: Training was interrupted"

        early_stop = parser.get_early_stop(std_log_data)

        self.assertTrue(early_stop)

    @mock.patch("builtins.open", new_callable=mock.mock_open)
    def test_get_consumption(self, mock_open):
        output_log_data = (
            "2022-11-14 15:44:48 - CarbonTracker: Actual consumption for 1 epoch(s):\n"
            "	Time:	0:02:22\n"
            "	Energy:	0.009417 kWh\n"
            "	CO2eq:	0.960490 g\n"
            "	This is equivalent to:\n"
            "	0.007977 km travelled by car\n"
        )

        mock_open.return_value.read.return_value = output_log_data

        actual, pred = parser.get_consumption(output_log_data)

        expected_actual = {
            "epochs": 1,
            "duration (s)": 142.0,
            "energy (kWh)": 0.009417,
            "co2eq (g)": 0.96049,
            "equivalents": {"km travelled by car": 0.007977},
        }
        expected_pred = None

        self.assertEqual(actual, expected_actual)
        self.assertEqual(pred, expected_pred)

    @mock.patch("builtins.open", new_callable=mock.mock_open)
    @mock.patch("os.listdir")
    @mock.patch("os.path.isfile")
    def test_parse_all_logs(self, mock_isfile, mock_listdir, mock_open):
        log_dir = "/path/to/logs"

        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker_output.log"),
            contents="output_log1 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker.log"),
            contents="std_log1 content",
        )

        mock_listdir.return_value = [
            "10151_2024-03-26T105926Z_carbontracker_output.log",
            "10151_2024-03-26T105926Z_carbontracker.log",
        ]
        mock_isfile.side_effect = lambda path: path.endswith(".log")
        mock_open.return_value.read.return_value = "content"

        logs = parser.parse_all_logs(log_dir)

        self.assertEqual(len(logs), 1)
        self.assertEqual(
            logs[0]["output_filename"],
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker_output.log"),
        )
        self.assertEqual(
            logs[0]["standard_filename"],
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker.log"),
        )

    @mock.patch("builtins.open", new_callable=mock.mock_open)
    @mock.patch("os.listdir")
    @mock.patch("os.path.isfile")
    @mock.patch("carbontracker.parser.get_devices")
    def test_parse_logs(self, mock_get_devices, mock_isfile, mock_listdir, mock_open):
        log_dir = "/path/to/logs"

        std_log_data = (
            "2022-11-14 15:44:48 - Average power usage (W) for gpu: [136.86084615]\n"
            "2022-11-14 15:44:48 - Average power usage (W) for cpu: [13.389104]\n"
            "2022-11-14 15:44:48 - Epoch 1:\nDuration: 0:02:21.90"
        )

        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker_output.log"),
            contents="output_log1 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker.log"),
            contents=std_log_data,
        )

        mock_isfile.side_effect = lambda path: path.endswith(".log")
        mock_open.return_value.read.return_value = std_log_data
        mock_get_devices.return_value = {
            "gpu": ["NVIDIA GeForce RTX 3060"],
            "cpu": ["cpu:0"],
        }

        components = parser.parse_logs(
            log_dir,
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker.log"),
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker_output.log"),
        )

        self.assertIn("gpu", components)
        self.assertIn("cpu", components)

    def test_get_avg_power_usages_none_power(self):
        std_log_data = "2022-11-14 15:44:48 - Average power usage (W) for gpu: None"

        avg_power_usages = parser.get_avg_power_usages(std_log_data)

        expected_avg_power_usages = {"gpu": [[0.0]]}

        self.assertEqual(avg_power_usages, expected_avg_power_usages)

    @mock.patch("os.listdir")
    @mock.patch("os.path.isfile")
    @mock.patch("os.path.getsize")
    def test_get_all_logs_mismatched_files(
        self, mock_getsize, mock_isfile, mock_listdir
    ):
        log_dir = "/path/to/logs"

        # Create three matching pairs of log files
        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker_output.log"),
            contents="output_log1 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker.log"),
            contents="std_log1 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "32487_2024-06-26T141608Z_carbontracker_output.log"),
            contents="output_log2 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "32487_2024-06-26T141608Z_carbontracker.log"),
            contents="std_log2 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "40793_2024-03-26T131535Z_carbontracker_output.log"),
            contents="output_log3 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "40793_2024-03-26T131535Z_carbontracker.log"),
            contents="std_log3 content",
        )

        # Add extra unmatched output log file
        self.fs.create_file(
            os.path.join(log_dir, "9803_2024-03-26T105836Z_carbontracker_output.log"),
            contents="output_log4 content",
        )

        mock_listdir.return_value = [
            "10151_2024-03-26T105926Z_carbontracker_output.log",
            "32487_2024-06-26T141608Z_carbontracker_output.log",
            "40793_2024-03-26T131535Z_carbontracker_output.log",
            "9803_2024-03-26T105836Z_carbontracker_output.log",
            "10151_2024-03-26T105926Z_carbontracker.log",
            "32487_2024-06-26T141608Z_carbontracker.log",
            "40793_2024-03-26T131535Z_carbontracker.log",
        ]

        mock_isfile.side_effect = lambda path: path.endswith(".log")
        mock_getsize.return_value = 100

        expected = [os.path.join(log_dir, f) for f in mock_listdir.return_value]
        self.assertTupleEqual(
            parser.get_all_logs(log_dir),
            (expected[:3], expected[4:]),
        )

    @mock.patch("os.listdir")
    @mock.patch("os.path.isfile")
    @mock.patch("os.path.getsize")
    def test_get_all_logs_mismatched_files_extra_std_log(
        self, mock_getsize, mock_isfile, mock_listdir
    ):
        log_dir = "/path/to/logs"

        # Create three matching pairs of log files
        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker_output.log"),
            contents="output_log1 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker.log"),
            contents="std_log1 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "32487_2024-06-26T141608Z_carbontracker_output.log"),
            contents="output_log2 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "32487_2024-06-26T141608Z_carbontracker.log"),
            contents="std_log2 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "40793_2024-03-26T131535Z_carbontracker_output.log"),
            contents="output_log3 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "40793_2024-03-26T131535Z_carbontracker.log"),
            contents="std_log3 content",
        )

        # Add extra unmatched std log file
        self.fs.create_file(
            os.path.join(log_dir, "9803_2024-03-26T105836Z_carbontracker.log"),
            contents="std_log4 content",
        )

        mock_listdir.return_value = [
            "10151_2024-03-26T105926Z_carbontracker_output.log",
            "32487_2024-06-26T141608Z_carbontracker_output.log",
            "40793_2024-03-26T131535Z_carbontracker_output.log",
            "10151_2024-03-26T105926Z_carbontracker.log",
            "32487_2024-06-26T141608Z_carbontracker.log",
            "40793_2024-03-26T131535Z_carbontracker.log",
            "9803_2024-03-26T105836Z_carbontracker.log",
        ]

        mock_isfile.side_effect = lambda path: path.endswith(".log")
        mock_getsize.return_value = 100

        expected = [os.path.join(log_dir, f) for f in mock_listdir.return_value]
        self.assertTupleEqual(
            parser.get_all_logs(log_dir), (expected[:3], expected[3:6])
        )

    @mock.patch("builtins.open", new_callable=mock.mock_open)
    def test_get_consumption_no_equivalents(self, mock_open):
        output_log_data = (
            "2022-11-14 15:44:48 - CarbonTracker: "
            "Actual consumption for 1 epoch(s):\n"
            "   Time:  0:02:22\n"
            "   Energy:    0.009417 kWh\n"
            "   CO2eq: 0.960490 g\n"
            "   This is equivalent to:\n"
        )

        mock_open.return_value.read.return_value = output_log_data

        actual, pred = parser.get_consumption(output_log_data)

        expected_actual = {
            "epochs": 1,
            "duration (s)": 142.0,
            "energy (kWh)": 0.009417,
            "co2eq (g)": 0.96049,
            "equivalents": {},
        }
        expected_pred = None

        self.assertEqual(actual, expected_actual)
        self.assertEqual(pred, expected_pred)

    def test_parse_logs_no_files(self):
        log_dir = "/logs"
        self.fs.create_file(log_dir + "/test_carbontracker.log")
        self.fs.create_file(log_dir + "/test_carbontracker_output.log")

        # No assertion - we are testing that no errors are raised
        parse_logs(log_dir)

    @mock.patch("builtins.open", new_callable=mock.mock_open)
    @mock.patch("carbontracker.parser.get_avg_power_usages", return_value={})
    @mock.patch("carbontracker.parser.get_devices")
    def test_parse_logs_consumption_no_power_usages(
        self, mock_get_devices, mock_power_usages, mock_open
    ):
        log_dir = "/logs"
        std_log_file = log_dir + "/test_carbontracker.log"
        output_log_file = log_dir + "/test_carbontracker_output.log"

        std_log_data = (
            "2022-11-14 15:42:43 - carbontracker version 1.1.6\n"
            "2022-11-14 15:42:43 - Only predicted and actual consumptions are multiplied by a PUE coefficient of 1.59 (Rhonda Ascierto, 2019, Uptime Institute Global Data Center Survey).\n"
            "2022-11-14 15:42:43 - The following components were found: GPU with device(s) NVIDIA GeForce RTX 3060. CPU with device(s) cpu:0.\n"
            "2022-11-14 15:42:43 - Monitoring thread started.\n"
            "2022-11-14 15:43:37 - Epoch 1:\n"
            "2022-11-14 15:43:37 - Duration: 0:00:53.25\n"
            "2022-11-14 15:43:37 - Average power usage (W) for gpu: [None]\n"
            "2022-11-14 15:43:37 - Average power usage (W) for cpu: [None]\n"
            "2022-11-14 15:43:37 - Carbon intensities (gCO2/kWh) fetched every 900 s at detected location Copenhagen, Capital Region, DK: [102.0]\n"
            "2022-11-14 15:43:37 - Average carbon intensity during training was 102.00 gCO2/kWh at detected location: Copenhagen, Capital Region, DK.\n"
            "2022-11-14 15:43:37 - Carbon intensity for the next 0:03:33 is predicted to be 102.00 gCO2/kWh at detected location: Copenhagen, Capital Region, DK.\n"
            "2022-11-14 15:43:37 - Monitoring thread ended.\n"
        )

        output_log_data = "No power usage data"

        self.fs.create_dir(log_dir)
        self.fs.create_file(std_log_file, contents=std_log_data)
        self.fs.create_file(output_log_file, contents=output_log_data)

        mock_open.return_value.read.return_value = std_log_data

        mock_get_devices.return_value = {
            "gpu": ["NVIDIA GeForce RTX 3060"],
            "cpu": ["cpu:0"],
        }

        components = parser.parse_logs(log_dir, std_log_file, output_log_file)

        expected_components = {
            "gpu": {
                "avg_power_usages (W)": None,
                "avg_energy_usages (J)": None,
                "epoch_durations (s)": [53.25],
                "devices": ["NVIDIA GeForce RTX 3060"],
            },
            "cpu": {
                "avg_power_usages (W)": None,
                "avg_energy_usages (J)": None,
                "epoch_durations (s)": [53.25],
                "devices": ["cpu:0"],
            },
        }

        self.assertEqual(components, expected_components)

    @mock.patch("re.search")
    def test_extract_measurements_no_match(self, mock_search):
        mock_search.return_value = None

        measurements = extract_measurements(None)

        self.assertIsNone(measurements)

    @mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_print_aggregate(self, mock_stdout):
        log_dir = "/logs"
        self.fs.create_file(log_dir + "/test_carbontracker.log")
        self.fs.create_file(log_dir + "/test_carbontracker_output.log")

        print_aggregate(log_dir)

        self.assertIn("Measured by carbontracker", mock_stdout.getvalue())

    @mock.patch("carbontracker.parser.get_all_logs")
    def test_aggregate_consumption_all_logs_none(self, mock_get_all_logs):
        log_dir = "/path/to/logs"
        mock_get_all_logs.return_value = ([], [])

        total_energy, total_co2eq, total_equivalents = parser.aggregate_consumption(
            log_dir
        )

        self.assertEqual(total_energy, 0)
        self.assertEqual(total_co2eq, 0)
        self.assertEqual(total_equivalents, {})

    @mock.patch("carbontracker.parser.get_all_logs")
    @mock.patch("carbontracker.parser.get_consumption")
    @mock.patch("carbontracker.parser.get_early_stop")
    def test_aggregate_consumption(
        self, mock_get_early_stop, mock_get_consumption, mock_get_all_logs
    ):
        log_dir = "/path/to/logs"
        output_log_path = "/path/to/logs/output_log1"
        std_log_path = "/path/to/logs/std_log1"

        self.fs.create_file(output_log_path, contents="output_log_content")
        self.fs.create_file(std_log_path, contents="std_log_content")

        mock_get_all_logs.return_value = ([output_log_path], [std_log_path])
        mock_get_consumption.return_value = (None, None)
        mock_get_early_stop.return_value = False

        with mock.patch(
            "builtins.open", mock.mock_open(read_data="mock_data")
        ) as mock_open:
            total_energy, total_co2eq, total_equivalents = parser.aggregate_consumption(
                log_dir
            )

        expected_total_energy = 0
        expected_total_co2eq = 0
        expected_total_equivalents = {}

        self.assertEqual(total_energy, expected_total_energy)
        self.assertEqual(total_co2eq, expected_total_co2eq)
        self.assertEqual(total_equivalents, expected_total_equivalents)

    @mock.patch("builtins.open", new_callable=mock.mock_open)
    @mock.patch("os.listdir")
    @mock.patch("os.path.isfile")
    def test_aggregate_consumption_actual(self, mock_isfile, mock_listdir, mock_open):
        log_dir = "/path/to/logs"

        output_log_content = (
            "2022-11-14 15:44:48 - CarbonTracker: Actual consumption for 1 epoch(s):\n"
            "	Time:	0:02:22\n"
            "	Energy:	0.009417 kWh\n"
            "	CO2eq:	0.960490 g\n"
            "	This is equivalent to:\n"
            "	0.007977 km travelled by car\n"
        )

        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker_output.log"),
            contents=output_log_content,
        )
        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker.log"),
            contents="std_log1 content",
        )

        mock_listdir.return_value = [
            "10151_2024-03-26T105926Z_carbontracker_output.log",
            "10151_2024-03-26T105926Z_carbontracker.log",
        ]
        mock_isfile.side_effect = lambda path: path.endswith(".log")
        mock_open.return_value.read.return_value = output_log_content

        total_energy, total_co2eq, total_equivalents = parser.aggregate_consumption(
            log_dir
        )

        self.assertEqual(total_energy, 0.009417)
        self.assertEqual(total_co2eq, 0.96049)
        self.assertEqual(total_equivalents, {"km travelled by car": 0.007977})

    @mock.patch("carbontracker.parser.get_all_logs")
    @mock.patch("carbontracker.parser.get_consumption")
    @mock.patch("carbontracker.parser.get_early_stop")
    def test_aggregate_consumption_both_none(
        self, mock_get_early_stop, mock_get_consumption, mock_get_all_logs
    ):
        log_dir = "/path/to/logs"
        output_log_path = "/path/to/logs/output_log1"
        std_log_path = "/path/to/logs/std_log1"

        self.fs.create_file(output_log_path, contents="output_log_content")
        self.fs.create_file(std_log_path, contents="std_log_content")

        mock_get_all_logs.return_value = ([output_log_path], [std_log_path])
        mock_get_consumption.return_value = (None, None)
        mock_get_early_stop.return_value = False

        total_energy, total_co2eq, total_equivalents = parser.aggregate_consumption(
            log_dir
        )

        expected_total_energy = 0
        expected_total_co2eq = 0
        expected_total_equivalents = {}

        self.assertEqual(total_energy, expected_total_energy)
        self.assertEqual(total_co2eq, expected_total_co2eq)
        self.assertEqual(total_equivalents, expected_total_equivalents)

    @mock.patch("carbontracker.parser.get_all_logs")
    @mock.patch("carbontracker.parser.get_consumption")
    @mock.patch("carbontracker.parser.get_early_stop")
    def test_aggregate_consumption_actual_none(
        self, mock_get_early_stop, mock_get_consumption, mock_get_all_logs
    ):
        log_dir = "/path/to/logs"
        output_log_path = "/path/to/logs/output_log1"
        std_log_path = "/path/to/logs/std_log1"

        self.fs.create_file(output_log_path, contents="output_log_content")
        self.fs.create_file(std_log_path, contents="std_log_content")

        mock_get_all_logs.return_value = ([output_log_path], [std_log_path])
        mock_get_consumption.return_value = (
            None,
            {"energy (kWh)": 1, "co2eq (g)": 2, "equivalents": {"km": 3}},
        )
        mock_get_early_stop.return_value = False

        total_energy, total_co2eq, total_equivalents = parser.aggregate_consumption(
            log_dir
        )

        expected_total_energy = 1
        expected_total_co2eq = 2
        expected_total_equivalents = {"km": 3}

        self.assertEqual(total_energy, expected_total_energy)
        self.assertEqual(total_co2eq, expected_total_co2eq)
        self.assertEqual(total_equivalents, expected_total_equivalents)

    @mock.patch("carbontracker.parser.get_all_logs")
    @mock.patch("carbontracker.parser.get_consumption")
    @mock.patch("carbontracker.parser.get_early_stop")
    def test_aggregate_consumption_pred_none(
        self, mock_get_early_stop, mock_get_consumption, mock_get_all_logs
    ):
        log_dir = "/path/to/logs"
        output_log_path = "/path/to/logs/output_log1"
        std_log_path = "/path/to/logs/std_log1"

        self.fs.create_file(output_log_path, contents="output_log_content")
        self.fs.create_file(std_log_path, contents="std_log_content")

        mock_get_all_logs.return_value = ([output_log_path], [std_log_path])
        mock_get_consumption.return_value = (
            {"energy (kWh)": 1, "co2eq (g)": 2, "equivalents": {"km": 3}},
            None,
        )
        mock_get_early_stop.return_value = False

        total_energy, total_co2eq, total_equivalents = parser.aggregate_consumption(
            log_dir
        )

        expected_total_energy = 1
        expected_total_co2eq = 2
        expected_total_equivalents = {"km": 3}

        self.assertEqual(total_energy, expected_total_energy)
        self.assertEqual(total_co2eq, expected_total_co2eq)
        self.assertEqual(total_equivalents, expected_total_equivalents)

    @mock.patch("carbontracker.parser.get_all_logs")
    @mock.patch("carbontracker.parser.get_consumption")
    @mock.patch("carbontracker.parser.get_early_stop")
    def test_aggregate_consumption_both_available(
        self, mock_get_early_stop, mock_get_consumption, mock_get_all_logs
    ):
        log_dir = "/path/to/logs"
        output_log_path = "/path/to/logs/output_log1"
        std_log_path = "/path/to/logs/std_log1"

        output_log_content = (
            "2022-11-14 15:42:43 - CarbonTracker: The following components were found: GPU with device(s) NVIDIA GeForce RTX 3060. CPU with device(s) cpu:0.\n"
            "2022-11-14 15:43:37 - CarbonTracker: \n"
            "Actual consumption for 1 epoch(s):\n"
            "    Time:    0:00:53\n"
            "    Energy:    0.003504 kWh\n"
            "    CO2eq:    0.357451 g\n"
            "    This is equivalent to:\n"
            "    0.002969 km travelled by car\n"
            "2022-11-14 15:43:37 - CarbonTracker: \n"
            "Predicted consumption for 4 epoch(s):\n"
            "    Time:    0:03:33\n"
            "    Energy:    0.014018 kWh\n"
            "    CO2eq:    1.429803 g\n"
            "    This is equivalent to:\n"
            "    0.011875 km travelled by car\n"
            "2022-11-14 15:43:37 - CarbonTracker: Finished monitoring."
        )
        std_log_content = "std_log_content"

        self.fs.create_file(output_log_path, contents=output_log_content)
        self.fs.create_file(std_log_path, contents=std_log_content)

        mock_get_all_logs.return_value = ([output_log_path], [std_log_path])
        mock_get_consumption.return_value = (
            {
                "epochs": 1,
                "duration (s)": 53,
                "energy (kWh)": None,
                "co2eq (g)": None,
                "equivalents": None,
            },
            {
                "epochs": 4,
                "duration (s)": 213,
                "energy (kWh)": 0.014018,
                "co2eq (g)": 1.429803,
                "equivalents": {"km travelled by car": 0.011875},
            },
        )
        mock_get_early_stop.return_value = False

        total_energy, total_co2eq, total_equivalents = parser.aggregate_consumption(
            log_dir
        )

        expected_total_energy = 0.014018
        expected_total_co2eq = 1.429803
        expected_total_equivalents = {"km travelled by car": 0.011875}

        self.assertEqual(total_energy, expected_total_energy)
        self.assertEqual(total_co2eq, expected_total_co2eq)
        self.assertEqual(total_equivalents, expected_total_equivalents)

    @mock.patch("carbontracker.parser.get_all_logs")
    @mock.patch("carbontracker.parser.get_consumption")
    @mock.patch("carbontracker.parser.get_early_stop")
    def test_aggregate_consumption_multiple_files(
        self, mock_get_early_stop, mock_get_consumption, mock_get_all_logs
    ):
        log_dir = "/path/to/logs"
        output_log_path1 = "/path/to/logs/output_log1"
        std_log_path1 = "/path/to/logs/std_log1"
        output_log_path2 = "/path/to/logs/output_log2"
        std_log_path2 = "/path/to/logs/std_log2"

        output_log_content1 = (
            "2022-11-14 15:42:43 - CarbonTracker: The following components were found: GPU with device(s) NVIDIA GeForce RTX 3060. CPU with device(s) cpu:0.\n"
            "2022-11-14 15:43:37 - CarbonTracker: \n"
            "Actual consumption for 1 epoch(s):\n"
            "    Time:    0:00:53\n"
            "    Energy:    0.003504 kWh\n"
            "    CO2eq:    0.357451 g\n"
            "    This is equivalent to:\n"
            "    0.002969 km travelled by car\n"
        )
        std_log_content1 = "std_log_content1"

        output_log_content2 = (
            "2022-11-14 15:44:48 - CarbonTracker: \n"
            "Actual consumption for 2 epoch(s):\n"
            "    Time:    0:01:45\n"
            "    Energy:    0.007008 kWh\n"
            "    CO2eq:    0.714902 g\n"
            "    This is equivalent to:\n"
            "    0.005937 km travelled by car\n"
        )
        std_log_content2 = "std_log_content2"

        self.fs.create_file(output_log_path1, contents=output_log_content1)
        self.fs.create_file(std_log_path1, contents=std_log_content1)
        self.fs.create_file(output_log_path2, contents=output_log_content2)
        self.fs.create_file(std_log_path2, contents=std_log_content2)

        mock_get_all_logs.return_value = (
            [output_log_path1, output_log_path2],
            [std_log_path1, std_log_path2],
        )
        mock_get_consumption.side_effect = [
            (
                {
                    "epochs": 1,
                    "duration (s)": 53,
                    "energy (kWh)": 0.003504,
                    "co2eq (g)": 0.357451,
                    "equivalents": {"km travelled by car": 0.002969},
                },
                None,
            ),
            (
                {
                    "epochs": 2,
                    "duration (s)": 105,
                    "energy (kWh)": 0.007008,
                    "co2eq (g)": 0.714902,
                    "equivalents": {"km travelled by car": 0.005937},
                },
                None,
            ),
        ]
        mock_get_early_stop.return_value = False

        total_energy, total_co2eq, total_equivalents = parser.aggregate_consumption(
            log_dir
        )

        expected_total_energy = 0.010512  # Sum of energy from both logs
        expected_total_co2eq = 1.072353
        expected_total_equivalents = {"km travelled by car": 0.008906}

        self.assertAlmostEqual(total_energy, expected_total_energy, places=6)
        self.assertAlmostEqual(total_co2eq, expected_total_co2eq, places=6)
        self.assertAlmostEqual(
            total_equivalents["km travelled by car"],
            expected_total_equivalents["km travelled by car"],
            places=6,
        )

    @mock.patch("carbontracker.parser.get_all_logs")
    @mock.patch("carbontracker.parser.get_consumption")
    @mock.patch("carbontracker.parser.get_early_stop")
    def test_aggregate_consumption_early_stop(
        self, mock_get_early_stop, mock_get_consumption, mock_get_all_logs
    ):
        log_dir = "/path/to/logs"
        output_log_path = "/path/to/logs/output_log1"
        std_log_path = "/path/to/logs/std_log1"

        output_log_content = (
            "2022-11-14 15:43:37 - CarbonTracker: \n"
            "Actual consumption for 1 epoch(s):\n"
            "    Time:    0:00:53\n"
            "    Energy:    0.003504 kWh\n"
            "    CO2eq:    0.357451 g\n"
            "    This is equivalent to:\n"
            "    0.002969 km travelled by car\n"
            "2022-11-14 15:43:37 - CarbonTracker: \n"
            "Predicted consumption for 1 epoch(s):\n"
            "    Time:    0:00:53\n"
            "    Energy:    0.003000 kWh\n"
            "    CO2eq:    0.305000 g\n"
            "    This is equivalent to:\n"
            "    0.002500 km travelled by car\n"
        )
        std_log_content = "std_log_content"

        self.fs.create_file(output_log_path, contents=output_log_content)
        self.fs.create_file(std_log_path, contents=std_log_content)

        mock_get_all_logs.return_value = ([output_log_path], [std_log_path])
        mock_get_consumption.return_value = (
            {
                "epochs": 1,
                "duration (s)": 53,
                "energy (kWh)": 0.003504,
                "co2eq (g)": 0.357451,
                "equivalents": {"km travelled by car": 0.002969},
            },
            {
                "epochs": 1,
                "duration (s)": 53,
                "energy (kWh)": 0.003000,
                "co2eq (g)": 0.305000,
                "equivalents": {"km travelled by car": 0.002500},
            },
        )
        mock_get_early_stop.return_value = True

        total_energy, total_co2eq, total_equivalents = parser.aggregate_consumption(
            log_dir
        )

        expected_total_energy = 0.003504  # Energy from actual
        expected_total_co2eq = 0.357451
        expected_total_equivalents = {"km travelled by car": 0.002969}

        self.assertAlmostEqual(total_energy, expected_total_energy, places=6)
        self.assertAlmostEqual(total_co2eq, expected_total_co2eq, places=6)
        self.assertAlmostEqual(
            total_equivalents["km travelled by car"],
            expected_total_equivalents["km travelled by car"],
            places=6,
        )

    def test_get_time_no_match(self):
        time_str = "Invalid time string"
        result = parser.get_time(time_str)
        assert result is None

    @mock.patch("builtins.print")
    @mock.patch(
        "carbontracker.parser.aggregate_consumption", return_value=(100.0, 50000.0, {})
    )
    def test_print_aggregate_empty_equivalents(
        self, mock_aggregate_consumption, mock_print
    ):
        log_dir = "/logs"
        print_aggregate(log_dir)
        mock_print.assert_called_once_with(
            "The training of models in this work is estimated to use 100.0000000000000000 kWh of electricity contributing to 50.0000000000000000 kg of CO2eq. Measured by carbontracker (https://github.com/lfwa/carbontracker)."
        )

    @mock.patch("builtins.print")
    @mock.patch(
        "carbontracker.parser.aggregate_consumption",
        return_value=(100.0, 5000.0, {"km travelled": 200.0}),
    )
    def test_print_aggregate_non_empty_equivalents(
        self, mock_aggregate_consumption, mock_print
    ):
        log_dir = "/logs"
        print_aggregate(log_dir)
        mock_print.assert_called_once_with(
            "The training of models in this work is estimated to use 100.0000000000000000 kWh of electricity contributing to 5.0000000000000000 kg of CO2eq. "
            "This is equivalent to 200.0000000000000000 km travelled. "
            "Measured by carbontracker (https://github.com/lfwa/carbontracker)."
        )

    def test_get_stats_no_equivalents(self):
        groups = ["group1", "group2", "10.5", "20.5"]
        energy, co2eq, equivalents = get_stats(groups)
        self.assertEqual(energy, 10.5)
        self.assertEqual(co2eq, 20.5)
        self.assertIsNone(equivalents)

    def test_parse_equivalents_value_error(self):
        lines = "not_a_float equivalent1\n10.5 equivalent2"
        equivalents = parse_equivalents(lines)
        self.assertEqual({"equivalent2": 10.5}, equivalents)

    @mock.patch("builtins.open", new_callable=mock.mock_open)
    @mock.patch("os.listdir")
    @mock.patch("os.path.isfile")
    @mock.patch("carbontracker.parser.get_devices")
    def test_parse_epoch_mismatch(
        self, mock_get_devices, mock_isfile, mock_listdir, mock_open
    ):
        log_dir = "/path/to/logs"

        std_log_data = (
            "2022-11-14 15:44:48 - Average power usage (W) for gpu: [136.86084615]\n"
            "2022-11-14 15:44:48 - Average power usage (W) for cpu: [13.389104]\n"
            "2022-11-14 15:44:48 - Average power usage (W) for gpu: [136.86084615]\n"
            "2022-11-14 15:44:48 - Average power usage (W) for cpu: [13.389104]\n"
            "2022-11-14 15:44:48 - Epoch 1:\nDuration: 0:02:21.90"
        )

        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker_output.log"),
            contents="output_log1 content",
        )
        self.fs.create_file(
            os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker.log"),
            contents=std_log_data,
        )

        mock_isfile.side_effect = lambda path: path.endswith(".log")
        mock_open.return_value.read.return_value = std_log_data
        mock_get_devices.return_value = {
            "gpu": ["NVIDIA GeForce RTX 3060"],
            "cpu": ["cpu:0"],
        }

        with self.assertRaises(exceptions.MismatchedEpochsError):
            components = parser.parse_logs(
                log_dir,
                os.path.join(log_dir, "10151_2024-03-26T105926Z_carbontracker.log"),
                os.path.join(
                    log_dir, "10151_2024-03-26T105926Z_carbontracker_output.log"
                ),
            )

    def test_parse_logs_mismatch(self):
        results = parser.get_avg_power_usages("2024-03-26 10:51:53 - Epoch 1:\n2024-03-26 10:51:53 - Duration: 0:00:00.00\n2024-03-26 10:51:53 - Average power usage (W) for cpu: None\n2024-03-26 10:51:53 - Average power usage (W) for gpu: None")
        self.assertEqual(results, {"cpu": [[0.0]], "gpu": [[0.0]]})


class NoComponentsAvailableError(Exception):
    def __init__(
        self,
        msg=(
            "No components were available. CarbonTracker supports Intel "
            "CPUs with the RAPL interface and NVIDIA GPUs."
        ),
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)


class UnitError(Exception):
    """Raised when the expected unit does not match the received unit."""

    def __init__(self, expected_unit, received_unit, message):
        self.expected_unit = expected_unit
        self.received_unit = received_unit
        self.message = message


class IntelRaplPermissionError(Exception):
    """Raised when an Intel RAPL permission error occurs."""

    pass


class GPUPowerUsageRetrievalError(Exception):
    """Raised when a GPU power usage retrieval error occurs."""

    pass


class CarbonIntensityFetcherError(Exception):
    pass


class IPLocationError(Exception):
    pass


class GPUError(Exception):
    pass


class CPUError(Exception):
    pass


class ComponentNameError(Exception):
    pass


class FetcherNameError(Exception):
    pass


class MismatchedLogFilesError(Exception):
    pass


class MismatchedEpochsError(Exception):
    pass



import subprocess
import unittest
from unittest import skipIf
from time import sleep
from unittest.mock import patch
from io import StringIO
import sys
from carbontracker import cli
import os

def mock_password_input(prompt):
    # Simulate password entry based on the prompt
    if "Password:" in prompt:
        return "your_password"
    else:
        # Handle other prompts or return None for unexpected prompts
        return None

@skipIf(os.environ.get('CI') == 'true', 'Skipped due to CI')
class TestCLI(unittest.TestCase):

    @patch("builtins.input", side_effect=mock_password_input)
    @patch("sys.argv", ["python -c 'print('Test')'", "--log_dir", "./test_logs"])
    def test_main_with_args(self, mock_input):
        sleep(2)
        captured_output = StringIO()
        sys.stdout = captured_output

        cli.main()
        self.assertIn("CarbonTracker: The following components", captured_output.getvalue())


    @patch("builtins.input", side_effect=mock_password_input)
    @patch("sys.argv", ["python -c 'print('Test')'"])
    def test_main_without_args(self, mock_input):
        sleep(2)
        captured_output = StringIO()
        sys.stdout = captured_output

        cli.main()
        self.assertIn("CarbonTracker: The following components", captured_output.getvalue())

    @patch("builtins.input", side_effect=mock_password_input)
    @patch("subprocess.run", autospec=True)
    @patch.object(sys, "argv", ["cli.py", "--log_dir", "./logs", "echo 'test'"])
    def test_main_with_remaining_args(self, mock_subprocess, mock_input):
        sleep(2)
        mock_subprocess.return_value.returncode = 0  # Simulate a successful command execution

        cli.main()
        mock_subprocess.assert_called_once_with(["echo 'test'"], check=True)

    @patch("builtins.input", side_effect=mock_password_input)
    @patch("subprocess.run", autospec=True)
    @patch.object(sys, "argv", ["cli.py", "--log_dir", "./logs", "echo 'test'"])
    def test_main_with_remaining_args_failure(self, mock_subprocess, mock_input):
        sleep(2)
        mock_subprocess.side_effect = subprocess.CalledProcessError(0, ["echo 'test'"])

        cli.main()
        mock_subprocess.assert_called_once_with(["echo 'test'"], check=True)

if __name__ == "__main__":
    unittest.main()



import unittest
from unittest import mock
import datetime
from carbontracker.emissions.intensity.fetchers import energidataservice
from carbontracker import exceptions


class TestEnergiDataService(unittest.TestCase):
    def setUp(self):
        self.fetcher = energidataservice.EnergiDataService()
        self.geocoder = mock.MagicMock()
        self.geocoder.country = "DK"

    def test_suitable(self):
        self.assertTrue(self.fetcher.suitable(self.geocoder))
        self.geocoder.country = "US"
        self.assertFalse(self.fetcher.suitable(self.geocoder))

    @mock.patch("requests.get")
    def test_carbon_intensity_no_time_dur(self, mock_get):
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.json.return_value = {
            "records": [{"CO2Emission": 1.0}, {"CO2Emission": 2.0}]
        }
        mock_get.return_value = mock_response
        result = self.fetcher.carbon_intensity(self.geocoder)

        self.assertEqual(result.carbon_intensity, 1.0)
        self.assertFalse(result.is_prediction)

    @mock.patch("requests.get")
    def test_carbon_intensity_with_time_dur(self, mock_get):
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.json.return_value = {
            "records": [
                {"CO2Emission": 1.0},
                {"CO2Emission": 2.0},
                {"CO2Emission": 3.0},
                {"CO2Emission": 4.0},
            ]
        }
        mock_get.return_value = mock_response

        result = self.fetcher.carbon_intensity(self.geocoder, time_dur=1800)

        self.assertEqual(result.carbon_intensity, 2.5)
        self.assertTrue(result.is_prediction)

    @mock.patch("requests.get")
    def test_nearest_5_min(self, mock_get):
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.json.return_value = {
            "records": [
                {"CO2Emission": 1.0},
                {"CO2Emission": 2.0},
                {"CO2Emission": 3.0},
                {"CO2Emission": 4.0},
            ]
        }
        mock_get.return_value = mock_response

        _result = self.fetcher.carbon_intensity(self.geocoder, time_dur=1800)

        now = datetime.datetime.now(datetime.timezone.utc)

        from_time = now - datetime.timedelta(
            minutes=now.minute % 5, seconds=now.second, microseconds=now.microsecond
        )
        to_time = from_time + datetime.timedelta(seconds=1800)

        # Format the from_time and to_time to strings
        date_format = "%Y-%m-%dT%H:%M"
        expected_from_time = from_time.strftime(date_format)
        expected_to_time = to_time.strftime(date_format)

        # Check that the mocked requests.get was called with the expected URL
        expected_url = f"https://api.energidataservice.dk/dataset/CO2Emis?start={expected_from_time}&end={expected_to_time}&limit=4"
        mock_get.assert_called_once_with(expected_url)

    @mock.patch("requests.get")
    def test_emission_current_response_not_ok(self, mock_get):
        mock_response = mock.MagicMock()
        mock_response.ok = False
        mock_response.json.return_value = {"error": "some error"}
        mock_get.return_value = mock_response

        with self.assertRaises(exceptions.CarbonIntensityFetcherError):
            self.fetcher._emission_current()

    @mock.patch("requests.get")
    def test_emission_prognosis_response_not_ok(self, mock_get):
        mock_response = mock.MagicMock()
        mock_response.ok = False
        mock_response.json.return_value = {"error": "some error"}
        mock_get.return_value = mock_response

        with self.assertRaises(exceptions.CarbonIntensityFetcherError):
            self.fetcher._emission_prognosis(time_dur=1800)



from unittest import TestCase, mock
from carbontracker.emissions.intensity.fetcher import IntensityFetcher


class TestIntensityFetcher(TestCase):
    def test_suitable_not_implemented(self):
        fetcher = IntensityFetcher()

        with self.assertRaises(NotImplementedError):
            fetcher.suitable(mock.MagicMock())

    def test_carbon_intensity_not_implemented(self):
        fetcher = IntensityFetcher()

        with self.assertRaises(NotImplementedError):
            fetcher.carbon_intensity(mock.MagicMock())



from unittest import TestCase, mock
import datetime
from carbontracker.emissions.intensity.fetchers import carbonintensitygb
from carbontracker import exceptions


class TestCarbonIntensityGB(TestCase):
    def setUp(self):
        self.fetcher = carbonintensitygb.CarbonIntensityGB()

    def test_suitable_with_gb_location(self):
        g_location = mock.MagicMock(country="GB")
        result = self.fetcher.suitable(g_location)
        self.assertTrue(result)

    def test_suitable_with_non_gb_location(self):
        g_location = mock.MagicMock(country="US")
        result = self.fetcher.suitable(g_location)
        self.assertFalse(result)

    @mock.patch("carbontracker.emissions.intensity.fetchers.carbonintensitygb.datetime")
    @mock.patch(
        "carbontracker.emissions.intensity.fetchers.carbonintensitygb.requests.get"
    )
    def test_carbon_intensity_gb_regional(self, mock_get, mock_datetime):
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.json.return_value = {
            "data": {
                "data": [
                    {
                        "from": "2023-05-20T00:00Z",
                        "to": "2023-05-20T01:00Z",
                        "intensity": {"forecast": 200},
                    },
                    {
                        "from": "2023-05-20T01:00Z",
                        "to": "2023-05-20T02:00Z",
                        "intensity": {"forecast": 300},
                    },
                ]
            }
        }
        mock_get.return_value = mock_response
        g_location = mock.MagicMock(postal="AB12 3CD")
        time_dur = 3600

        # Patch datetime.now to return a fixed value
        mock_datetime.datetime.now.return_value = datetime.datetime(2023, 5, 20, 0, 0)

        # Patch datetime.timedelta to return a fixed value
        mock_datetime.timedelta().__radd__().strftime.return_value = "2023-05-20T01:00Z"

        result = self.fetcher._carbon_intensity_gb_regional(g_location.postal, time_dur)

        from_str = "2023-05-20T00:00Z"
        to_str = "2023-05-20T01:00Z"

        mock_get.assert_called_once_with(
            f"https://api.carbonintensity.org.uk/regional/intensity/{from_str}/{to_str}/postcode/AB12 3CD"
        )
        self.assertEqual(result, 250)

    @mock.patch(
        "carbontracker.emissions.intensity.fetchers.carbonintensitygb.requests.get"
    )
    def test_carbon_intensity_gb_regional_with_error_response(self, mock_get):
        mock_response = mock.MagicMock()
        mock_response.ok = False
        mock_response.json.return_value = {"error": "Invalid postcode"}
        mock_get.return_value = mock_response
        g_location = mock.MagicMock(postal="AB12 3CD")
        time_dur = 3600

        with self.assertRaises(exceptions.CarbonIntensityFetcherError):
            self.fetcher._carbon_intensity_gb_regional(g_location.postal, time_dur)

    @mock.patch("carbontracker.emissions.intensity.fetchers.carbonintensitygb.datetime")
    @mock.patch(
        "carbontracker.emissions.intensity.fetchers.carbonintensitygb.requests.get"
    )
    def test_carbon_intensity_gb_national(self, mock_get, mock_datetime):
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.json.return_value = {
            "data": [
                {
                    "intensity": {"forecast": 250},
                }
            ]
        }
        mock_get.return_value = mock_response
        time_dur = 3600

        # Set fixed values for from_str and to_str
        from_str = "2023-05-20T00:00Z"
        to_str = "2023-05-20T01:00Z"

        # Patch datetime.now to return a fixed value
        mock_datetime.datetime.now.return_value = datetime.datetime(2023, 5, 20, 0, 0)

        # Patch datetime.timedelta to return a fixed value
        mock_datetime.timedelta().__radd__().strftime.return_value = to_str

        result = self.fetcher._carbon_intensity_gb_national(time_dur)

        mock_get.assert_called_once_with(
            f"https://api.carbonintensity.org.uk/intensity/{from_str}/{to_str}"
        )
        self.assertEqual(result, 250)

    @mock.patch(
        "carbontracker.emissions.intensity.fetchers.carbonintensitygb.requests.get"
    )
    def test_carbon_intensity_gb_national_with_error_response(self, mock_get):
        mock_response = mock.MagicMock()
        mock_response.ok = False
        mock_response.json.return_value = {"error": "Invalid request"}
        mock_get.return_value = mock_response
        time_dur = 3600

        with self.assertRaises(exceptions.CarbonIntensityFetcherError):
            self.fetcher._carbon_intensity_gb_national(time_dur)

    def test_time_from_to_str(self):
        time_dur = 3600
        time_from = datetime.datetime.now(datetime.timezone.utc)
        time_to = time_from + datetime.timedelta(seconds=time_dur)
        from_str = time_from.strftime("%Y-%m-%dT%H:%MZ")
        to_str = time_to.strftime("%Y-%m-%dT%H:%MZ")

        result = self.fetcher._time_from_to_str(time_dur)

        self.assertEqual(result, (from_str, to_str))

    @mock.patch(
        "carbontracker.emissions.intensity.fetchers.carbonintensitygb.requests.get"
    )
    def test_carbon_intensity_with_postal(self, mock_get):
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.json.return_value = {"data": [{"intensity": {"forecast": 250}}]}
        mock_get.return_value = mock_response

        mock_get.__getitem__.return_value = mock_response
        g_location = mock.MagicMock(postal="AB12 3CD", country="GB")
        time_dur = 3600

        carbon_intensity_obj = self.fetcher.carbon_intensity(g_location, time_dur)
        self.assertEqual(carbon_intensity_obj.carbon_intensity, 250)
        self.assertEqual(carbon_intensity_obj.is_prediction, True)

    @mock.patch(
        "carbontracker.emissions.intensity.fetchers.carbonintensitygb.requests.get"
    )
    def test_carbon_intensity_without_postal(self, mock_get):
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.json.return_value = {"data": [{"intensity": {"forecast": 250}}]}
        mock_get.return_value = mock_response

        g_location = mock.MagicMock(country="GB")
        time_dur = 3600

        carbon_intensity_obj = self.fetcher.carbon_intensity(g_location, time_dur)
        self.assertEqual(carbon_intensity_obj.carbon_intensity, 250)
        self.assertEqual(carbon_intensity_obj.is_prediction, True)

    @mock.patch(
        "carbontracker.emissions.intensity.fetchers.carbonintensitygb.requests.get"
    )
    def test_carbon_intensity_gb_regional_without_time_dur(self, mock_get):
        mock_response = mock.MagicMock()
        mock_response.ok = True
        mock_response.json.return_value = {"data": [{"intensity": {"forecast": 250}}]}
        mock_get.return_value = mock_response

        g_location = mock.MagicMock(postal="AB12 3CD", country="GB")

        carbon_intensity_obj = self.fetcher.carbon_intensity(g_location, None)
        self.assertEqual(carbon_intensity_obj.carbon_intensity, 250)



import unittest
from unittest.mock import patch
from carbontracker.components.apple_silicon.powermetrics import (
    AppleSiliconCPU,
    AppleSiliconGPU,
    PowerMetricsUnified,
)


class TestAppleSiliconCPU(unittest.TestCase):
    def setUp(self):
        self.cpu_handler = AppleSiliconCPU(pids=[], devices_by_pid=False)
        self.cpu_handler.init()

    def test_shutdown(self):
        self.cpu_handler.shutdown()

    @patch("platform.system", return_value="Darwin")
    def test_available_darwin(self, mock_platform):
        self.assertTrue(self.cpu_handler.available())

    @patch("platform.system", return_value="AlienOS")
    def test_available_not_darwin(self, mock_platform):
        self.assertFalse(self.cpu_handler.available())

    def test_devices(self):
        self.assertEqual(self.cpu_handler.devices(), ["CPU"])

    @patch(
        "carbontracker.components.apple_silicon.powermetrics.PowerMetricsUnified.get_output",
        return_value="CPU Power: 1000 mW",
    )
    def test_power_usage_with_match(self, mock_get_output):
        self.assertEqual(self.cpu_handler.power_usage(), [1.0])

    @patch(
        "carbontracker.components.apple_silicon.powermetrics.PowerMetricsUnified.get_output",
        return_value="No CPU Power data",
    )
    def test_power_usage_no_match(self, mock_get_output):
        self.assertEqual(self.cpu_handler.power_usage(), [0.0])


class TestAppleSiliconGPU(unittest.TestCase):
    def setUp(self):
        self.gpu_handler = AppleSiliconGPU(pids=[], devices_by_pid=False)
        self.gpu_handler.init()

    @patch("platform.system", return_value="Darwin")
    def test_available_darwin(self, mock_platform):
        self.assertTrue(self.gpu_handler.available())

    @patch("platform.system", return_value="Windows")
    def test_available_not_darwin(self, mock_platform):
        self.assertFalse(self.gpu_handler.available())

    def test_devices(self):
        self.assertEqual(self.gpu_handler.devices(), ["GPU", "ANE"])

    @patch(
        "carbontracker.components.apple_silicon.powermetrics.PowerMetricsUnified.get_output",
        return_value="GPU Power: 500 mW\nANE Power: 300 mW",
    )
    def test_power_usage_with_match(self, mock_get_output):
        self.assertEqual(len(self.gpu_handler.power_usage()), 1)
        self.assertAlmostEqual(self.gpu_handler.power_usage()[0], 0.8, places=2)

    @patch(
        "carbontracker.components.apple_silicon.powermetrics.PowerMetricsUnified.get_output",
        return_value="No GPU Power data",
    )
    def test_power_usage_no_match(self, mock_get_output):
        self.assertEqual(self.gpu_handler.power_usage(), [0.0])


class TestPowerMetricsUnified(unittest.TestCase):
    @patch("subprocess.check_output", return_value="Sample Output")
    @patch("time.time", side_effect=[100, 101, 102, 200, 202])
    def test_get_output_with_actual_call(self, mock_time, mock_check_output):
        # First call - should call subprocess
        output1 = PowerMetricsUnified.get_output()

        # Second call - should use cached output
        output2 = PowerMetricsUnified.get_output()
        self.assertIsNotNone(PowerMetricsUnified._last_updated)
        if PowerMetricsUnified._last_updated is None:
            self.fail()
        # Advance time to invalidate cache
        PowerMetricsUnified._last_updated -= 2

        # Third call - should call subprocess again
        output3 = PowerMetricsUnified.get_output()

        self.assertEqual(mock_check_output.call_count, 2)
        self.assertEqual(output1, "Sample Output")
        self.assertEqual(output2, "Sample Output")
        self.assertEqual(output3, "Sample Output")


class TestAppleSiliconGPUPowerUsage(unittest.TestCase):
    def setUp(self):
        self.gpu_handler = AppleSiliconGPU(pids=[], devices_by_pid=False)
        self.gpu_handler.init()

    @patch(
        "carbontracker.components.apple_silicon.powermetrics.PowerMetricsUnified.get_output",
        return_value="GPU Power: 500 mW\nANE Power: 300 mW",
    )
    def test_power_usage_with_match(self, mock_get_output):
        self.assertEqual(len(self.gpu_handler.power_usage()), 1)
        self.assertAlmostEqual(self.gpu_handler.power_usage()[0], 0.8, places=2)

    @patch(
        "carbontracker.components.apple_silicon.powermetrics.PowerMetricsUnified.get_output",
        return_value="No GPU Power data",
    )
    def test_power_usage_no_match(self, mock_get_output):
        self.assertEqual(self.gpu_handler.power_usage(), [0.0])


if __name__ == "__main__":
    unittest.main()



import os
import threading
import time
import traceback
import unittest
from unittest import mock, skipIf
from unittest.mock import Mock, patch, MagicMock
from threading import Event
from typing import List, Any
import numpy as np

from carbontracker import exceptions, constants
from carbontracker.tracker import (
    CarbonIntensityThread,
    CarbonTrackerThread,
    CarbonTracker,
)
from carbontracker.components.component import Component
from carbontracker.components.gpu import nvidia
from carbontracker.components.cpu import intel


class TestCarbonIntensityThread(unittest.TestCase):
    def setUp(self):
        self.logger = Mock()
        self.stop_event = Event()

    def test_init(self):
        thread = CarbonIntensityThread(self.logger, self.stop_event)
        self.assertEqual(thread.name, "CarbonIntensityThread")
        self.assertEqual(thread.daemon, True)

    @patch("carbontracker.tracker.intensity")
    def test_fetch_carbon_intensity_success(self, mock_intensity):
        mock_intensity.carbon_intensity.return_value.success = True
        mock_intensity.carbon_intensity.return_value.carbon_intensity = 10.5

        thread = CarbonIntensityThread(self.logger, self.stop_event)
        thread._fetch_carbon_intensity()

        self.assertEqual(thread.carbon_intensities[0].carbon_intensity, 10.5)

    @patch("carbontracker.tracker.intensity")
    def test_fetch_carbon_intensity_failure(self, mock_intensity):
        mock_intensity.carbon_intensity.return_value.success = False

        thread = CarbonIntensityThread(self.logger, self.stop_event)
        thread._fetch_carbon_intensity()

        self.assertEqual(len(thread.carbon_intensities), 0)

    @patch("carbontracker.tracker.intensity.CarbonIntensity")
    @patch("carbontracker.tracker.intensity")
    def test_predict_carbon_intensity(self, mock_intensity, mock_carbon_intensity):
        mock_ci_return = Mock()
        mock_ci_return.success = True
        mock_ci_return.carbon_intensity = 10.5

        mock_intensity.carbon_intensity.return_value = mock_ci_return

        mock_carbon_intensity.return_value = mock_ci_return

        thread = CarbonIntensityThread(self.logger, self.stop_event)
        thread.carbon_intensities.append(mock_ci_return)

        pred_time_dur = 1000
        ci = thread.predict_carbon_intensity(pred_time_dur)

        self.assertEqual(ci.carbon_intensity, 10.5)
        mock_intensity.set_carbon_intensity_message.assert_called_with(
            ci, pred_time_dur
        )
        self.logger.info.assert_called()
        self.logger.output.assert_called()

    @patch("carbontracker.tracker.intensity.CarbonIntensity")
    @patch("carbontracker.tracker.intensity")
    def test_average_carbon_intensity(self, mock_intensity, mock_carbon_intensity):
        mock_ci_return = Mock()
        mock_ci_return.success = True
        mock_ci_return.carbon_intensity = 10.5
        mock_ci_return.address = "test_address"

        mock_intensity.carbon_intensity.return_value = mock_ci_return

        mock_carbon_intensity.return_value = mock_ci_return

        thread = CarbonIntensityThread(self.logger, self.stop_event)
        thread.carbon_intensities.append(mock_ci_return)

        avg_ci = thread.average_carbon_intensity()

        self.assertEqual(avg_ci.carbon_intensity, 10.5)

    @patch("carbontracker.tracker.CarbonIntensityThread._fetch_carbon_intensity")
    def test_run_with_fetch_exception(self, mock_fetch_carbon_intensity):
        mock_fetch_carbon_intensity.side_effect = Exception("Test exception")
        mock_stop_event = MagicMock()
        mock_logger = MagicMock()

        thread = CarbonIntensityThread(mock_logger, mock_stop_event)
        thread.run()

        assert mock_logger.err_warn.called

    @patch("carbontracker.tracker.CarbonIntensityThread._fetch_carbon_intensity")
    def test_run_with_multiple_fetch_calls(self, mock_fetch_carbon_intensity):
        update_interval = 0.001
        wait_duration = 0.5

        mock_logger = MagicMock()
        stop_event = threading.Event()
        CarbonIntensityThread(mock_logger, stop_event, update_interval)
        time.sleep(wait_duration)

        assert mock_fetch_carbon_intensity.call_count > 1

    @patch("carbontracker.tracker.intensity.carbon_intensity")
    def test_average_carbon_intensity_empty_intensities(self, mock_carbon_intensity):
        mock_logger = MagicMock()
        stop_event = threading.Event()

        thread = CarbonIntensityThread(mock_logger, stop_event)
        thread.carbon_intensities = []
        thread.average_carbon_intensity()

        assert len(thread.carbon_intensities) == 1


class TestCarbonTrackerThread(unittest.TestCase):
    def setUp(self):
        self.mock_components: List[Any] = [
            MagicMock(name="Component1"),
            MagicMock(name="Component2"),
        ]

        for component in self.mock_components:
            component.available.return_value = True

        self.mock_logger = MagicMock(name="Logger")
        self.mock_delete = MagicMock(name="Delete")

        self.thread = CarbonTrackerThread(
            self.mock_components,
            self.mock_logger,
            False,
            self.mock_delete,
            update_interval=0.1,
        )

    def tearDown(self):
        self.thread.running = False
        self.thread.epoch_counter = 0
        self.thread.epoch_times = []

    def test_stop_tracker(self):
        self.thread.running = True
        self.thread.stop()

        self.assertFalse(self.thread.running)

        # assert_any_call because different log statements races in Python 3.11 in Github Actions
        self.mock_logger.info.assert_any_call("Monitoring thread ended.")
        self.mock_logger.output.assert_called_with(
            "Finished monitoring.", verbose_level=1
        )

    def test_stop_tracker_not_running(self):
        self.thread.running = False
        result = self.thread.stop()

        assert result is None

    @patch(
        "carbontracker.components.component.component_names",
        return_value=["gpu", "cpu"],
    )
    @patch(
        "carbontracker.components.component.handlers_by_name",
        return_value=[nvidia.NvidiaGPU, intel.IntelCPU],
    )
    def test_run_and_measure(self, mock_component_names, mock_handlers_by_name):
        self.thread.epoch_start()

        time.sleep(0.4)

        self.thread.epoch_end()
        for component in self.mock_components:
            component.collect_power_usage.assert_called_with(self.thread.epoch_counter)

    def test_init(self):
        mock_components: List[Component] = [
            MagicMock(name="Component1"),
            MagicMock(name="Component2"),
        ]
        mock_logger = MagicMock(name="Logger")
        mock_delete = MagicMock(name="Delete")

        thread = CarbonTrackerThread(mock_components, mock_logger, False, mock_delete)

        self.assertEqual(thread.components, mock_components)
        self.assertEqual(thread.logger, mock_logger)
        self.assertTrue(thread.running)
        self.assertEqual(thread.epoch_counter, 0)
        self.assertEqual(thread.epoch_times, [])
        self.assertEqual(thread.running, True)
        self.assertEqual(thread.daemon, True)

    def test_run_with_exception_ignore_errors(self):
        self.thread._components_remove_unavailable = MagicMock()
        self.thread._components_remove_unavailable.return_value = self.mock_components

        self.thread._components_init = MagicMock()
        self.thread._log_components_info = MagicMock()
        self.thread._components_shutdown = MagicMock()
        self.thread.ignore_errors = True

        self.thread._collect_measurements = MagicMock(
            side_effect=Exception("Mocked exception")
        )

        self.thread.logger.err_critical = MagicMock()
        self.thread.logger.output = MagicMock()

        os._exit = MagicMock()

        self.thread.running = True

        time.sleep(0.2)

        self.assertFalse(os._exit.called)

    def test_epoch_start(self):
        self.thread.epoch_counter = 0

        self.thread.epoch_start()

        self.assertEqual(self.thread.epoch_counter, 1)
        self.assertIsNotNone(self.thread.cur_epoch_time)

    def test_epoch_end(self):
        self.thread.cur_epoch_time = (
            time.time() - 1
        )  # Set a non-zero value for cur_epoch_time

        self.thread.epoch_end()
        time.sleep(0.2)

        self.assertTrue(self.thread.epoch_times)
        self.assertAlmostEqual(self.thread.epoch_times[-1], 1, delta=0.1)

    def test_epoch_end_too_short(self):
        mock_component: Any = MagicMock(name="Component")
        mock_component.power_usages = []

        self.thread.components = [mock_component]

        self.thread.cur_epoch_time = time.time()

        self.thread.epoch_end()

        self.assertTrue(self.thread.epoch_times)
        self.assertIsNotNone(self.thread.epoch_times[-1])
        self.mock_logger.err_warn.assert_called_with(
            "Epoch duration is too short for a measurement to be collected."
        )

    def test_no_components_available(self):
        self.thread.components = []

        with self.assertRaises(exceptions.NoComponentsAvailableError):
            self.thread.begin()

    def test_total_energy_per_epoch(self):
        mock_component1: Any = MagicMock(name="Component1")
        mock_component1.energy_usage.return_value = np.array([1.0, 2.0, 3.0])
        mock_component2: Any = MagicMock(name="Component2")
        mock_component2.energy_usage.return_value = np.array([2.0, 3.0, 4.0])

        self.thread.components = [mock_component1, mock_component2]

        self.thread.epoch_times = [1.0, 1.0, 1.0]

        total_energy = self.thread.total_energy_per_epoch()

        expected_total_energy = np.array([3.0, 5.0, 7.0]) * constants.PUE_2023
        np.testing.assert_array_equal(total_energy, expected_total_energy)

    @mock.patch("os._exit")
    def test_handle_error_ignore(self, mock_os_exit):
        self.thread.ignore_errors = True
        error = Exception("Test error")
        expected_err_str = f"Ignored error: {traceback.format_exc()}Continued training without monitoring..."

        self.thread._handle_error(error)

        self.mock_logger.err_critical.assert_called_with(expected_err_str)
        self.mock_logger.output.assert_called_with(expected_err_str)
        self.thread.delete.assert_called()
        mock_os_exit.assert_not_called()

    @mock.patch("os._exit")
    def test_handle_error_no_ignore_errors(self, mock_os_exit):
        self.thread.ignore_errors = False
        self.thread.logger = self.mock_logger
        self.thread._handle_error(Exception("Test exception"))

        self.mock_logger.err_critical.assert_called()
        self.mock_logger.output.assert_called()

        mock_os_exit.assert_called_with(70)

    @mock.patch("carbontracker.tracker.CarbonTrackerThread._handle_error")
    def test_run_exception_handling(self, mock_handle_error):
        mock_wait = mock.MagicMock()
        mock_wait.side_effect = Exception("Test exception")

        self.thread.measuring_event.wait = mock_wait
        self.thread.run()

        mock_handle_error.assert_called()


class TestCarbonTracker(unittest.TestCase):
    def setUp(self):
        self.mock_logger = MagicMock()
        self.mock_tracker_thread = MagicMock()
        self.mock_intensity_thread = MagicMock()

        with patch(
            "carbontracker.tracker.CarbonTrackerThread",
            return_value=self.mock_tracker_thread,
        ), patch(
            "carbontracker.tracker.CarbonIntensityThread",
            return_value=self.mock_intensity_thread,
        ), patch(
            "carbontracker.tracker.loggerutil.Logger", return_value=self.mock_logger
        ), patch(
            "carbontracker.tracker.CarbonTracker._output_actual"
        ) as self.mock_output_actual, patch(
            "carbontracker.tracker.CarbonTracker._delete"
        ) as self.mock_delete:
            self.tracker = CarbonTracker(
                epochs=5,
                epochs_before_pred=1,
                monitor_epochs=3,
                update_interval=10,
                interpretable=True,
                stop_and_confirm=True,
                ignore_errors=False,
                components="all",
                devices_by_pid=False,
                log_dir=None,
                log_file_prefix="",
                verbose=1,
                decimal_precision=6,
            )

    def tearDown(self):
        self.mock_logger = None
        self.mock_intensity_thread = None
        self.mock_tracker_thread = None
        self.tracker = None

    def test_epoch_start_increments_epoch_counter_and_starts_measurement(self):
        assert self.tracker is not None
        assert self.mock_tracker_thread is not None
        initial_epoch_counter = self.tracker.epoch_counter
        self.tracker.epoch_start()
        self.assertEqual(self.tracker.epoch_counter, initial_epoch_counter + 1)
        self.assertTrue(self.mock_tracker_thread.measuring_event.is_set())

    def test_check_input_yes(self):
        with patch("builtins.input", return_value="y"):
            assert self.tracker is not None
            assert self.mock_logger is not None
            self.tracker._check_input("y")
            self.mock_logger.output.assert_called_with("Continuing...")

    def test_check_input_no(self):
        assert self.tracker is not None
        with patch("builtins.input", return_value="n"):
            with self.assertRaises(SystemExit):
                self.tracker._check_input("n")

    @patch("carbontracker.tracker.CarbonTracker._check_input")
    def test_user_query(self, mock_check_input):
        assert self.tracker is not None
        with patch("builtins.input", return_value="y"), patch.object(
            self.tracker.logger, "output"
        ) as mock_logger_output:
            self.tracker._user_query()
            mock_logger_output.assert_called_once_with("Continue training (y/n)?")

        mock_check_input.assert_called_once()

    def test_check_input_invalid(self):
        assert self.tracker is not None
        assert self.mock_logger is not None
        with patch("builtins.input", side_effect=["a", "y"]):
            self.tracker._check_input("a")
            self.mock_logger.output.assert_any_call(
                "Input not recognized. Try again (y/n):"
            )
            self.tracker._check_input("y")
            self.mock_logger.output.assert_any_call("Continuing...")

    def test_delete(self):
        assert self.tracker is not None
        assert self.mock_tracker_thread is not None
        self.tracker._delete()
        self.mock_tracker_thread.stop.assert_called_once()
        self.assertTrue(self.tracker.deleted)

    @patch("carbontracker.tracker.psutil.Process")
    def test_get_pids(self, mock_process):
        assert self.tracker is not None
        mock_process.return_value.pid = 1234
        mock_process.return_value.children.return_value = [MagicMock(pid=5678)]
        pids = self.tracker._get_pids()
        self.assertEqual(pids, [1234, 5678])

    def test_stop_when_already_deleted(self):
        """Test the stop method when the tracker has already been marked as deleted."""
        assert self.tracker is not None
        assert self.mock_logger is not None
        self.tracker.deleted = True

        self.tracker.stop()

        self.mock_logger.info.assert_not_called()
        self.mock_output_actual.assert_not_called()
        self.mock_delete.assert_not_called()

    @patch("carbontracker.tracker.CarbonTracker._output_actual")
    def test_stop_behavior(self, mock_output_actual):
        assert self.tracker is not None
        self.assertFalse(self.tracker.deleted)

        initial_epoch_counter = 2
        self.tracker.epoch_counter = initial_epoch_counter
        self.tracker.stop()

        expected_epoch_counter = initial_epoch_counter - 1
        self.assertEqual(
            self.tracker.epoch_counter,
            expected_epoch_counter,
            "Epoch counter should be decremented by 1.",
        )

        mock_output_actual.assert_called_once()

        self.assertTrue(
            self.tracker.deleted,
            "Tracker should be marked as deleted after stop is called.",
        )

    def test_epoch_end_when_deleted(self):
        assert self.tracker is not None
        assert self.mock_tracker_thread is not None
        self.tracker.deleted = True
        self.tracker.epoch_end()
        self.mock_tracker_thread.epoch_end.assert_not_called()

    @patch("carbontracker.tracker.CarbonTracker._output_actual", autospec=True)
    @patch("carbontracker.tracker.CarbonTracker._delete", autospec=True)
    def test_epoch_end_output_actual_and_delete(self, mock_delete, mock_output_actual):
        assert self.tracker is not None
        self.tracker.epoch_counter = self.tracker.monitor_epochs
        self.tracker.epoch_end()

        mock_output_actual.assert_called_once()
        mock_delete.assert_called_once()

    @patch("carbontracker.tracker.CarbonTracker._output_pred", autospec=True)
    @patch("carbontracker.tracker.CarbonTracker._user_query", autospec=True)
    def test_epoch_end_output_pred_and_user_query(
        self, mock_user_query, mock_output_pred
    ):
        assert self.tracker is not None
        self.tracker.epoch_counter = self.tracker.epochs_before_pred
        self.tracker.epoch_end()

        mock_output_pred.assert_called_once()
        mock_user_query.assert_called_once()

    @patch("carbontracker.tracker.CarbonTracker._handle_error", autospec=True)
    def test_epoch_end_exception_handling(self, mock_handle_error):
        assert self.tracker is not None
        assert self.mock_tracker_thread is not None
        self.mock_tracker_thread.epoch_end.side_effect = Exception("Test Exception")
        self.tracker.epoch_end()

        mock_handle_error.assert_called_once()

    def test_invalid_monitor_epochs_value(self):
        with self.assertRaises(ValueError):
            CarbonTracker(
                epochs=5,
                monitor_epochs=0,  # Invalid value
                epochs_before_pred=2,
                update_interval=10,
                interpretable=True,
                stop_and_confirm=False,
                ignore_errors=False,
                components="all",
                devices_by_pid=False,
                log_dir=None,
                log_file_prefix="",
                verbose=1,
                decimal_precision=6,
            )

    def test_invalid_monitor_epochs_less_than_epochs_before_pred(self):
        with self.assertRaises(ValueError):
            CarbonTracker(
                epochs=5,
                monitor_epochs=1,
                epochs_before_pred=3,
                update_interval=10,
                interpretable=True,
                stop_and_confirm=False,
                ignore_errors=False,
                components="all",
                devices_by_pid=False,
                log_dir=None,
                log_file_prefix="",
                verbose=1,
                decimal_precision=6,
            )

    @patch("carbontracker.tracker.CarbonTracker._handle_error")
    def test_epoch_start_deleted(self, mock_handle_error):
        assert self.tracker is not None
        self.tracker.deleted = True
        self.tracker.epoch_start()

        self.assertEqual(self.tracker.epoch_counter, 0)

        mock_handle_error.assert_not_called()

    @skipIf(os.environ.get("CI") == "true", "Skipped due to CI")
    @patch("carbontracker.tracker.CarbonTrackerThread.epoch_start")
    @patch("carbontracker.tracker.CarbonTracker._handle_error")
    def test_epoch_start_exception(
        self, mock_handle_error, mock_tracker_thread_epoch_start
    ):
        tracker = CarbonTracker(
            epochs=5,
            epochs_before_pred=1,
            monitor_epochs=3,
            update_interval=10,
            interpretable=True,
            stop_and_confirm=True,
            ignore_errors=False,
            components="all",
            devices_by_pid=False,
            log_dir=None,
            log_file_prefix="",
            verbose=1,
            decimal_precision=6,
        )

        tracker.deleted = False
        mock_tracker_thread_epoch_start.side_effect = Exception("Test Exception")
        tracker.epoch_start()
        self.assertEqual(tracker.epoch_counter, 0)

        mock_tracker_thread_epoch_start.assert_called_once()
        mock_handle_error.assert_called_once()

    def test_handle_error_ignore_errors(self):
        assert self.tracker is not None
        assert self.mock_logger is not None
        self.tracker.ignore_errors = True
        self.tracker._handle_error(Exception("Test exception"))
        self.mock_logger.err_critical.assert_called_once()

    def test_handle_error_no_ignore_errors(self):
        assert self.tracker is not None
        self.tracker.ignore_errors = False
        with self.assertRaises(SystemExit):
            self.tracker._handle_error(Exception("Test exception"))

    @skipIf(os.environ.get("CI") == "true", "Skipped due to CI")
    @patch(
        "carbontracker.emissions.intensity.fetchers.electricitymaps.ElectricityMap.set_api_key"
    )
    def test_set_api_keys_electricitymaps(self, mock_set_api_key):
        tracker = CarbonTracker(epochs=1)
        api_dict = {"ElectricityMaps": "mock_api_key"}
        tracker.set_api_keys(api_dict)

        mock_set_api_key.assert_called_once_with("mock_api_key")

    @skipIf(os.environ.get("CI") == "true", "Skipped due to CI")
    @patch("carbontracker.tracker.CarbonTracker.set_api_keys")
    def test_carbontracker_api_key(self, mock_set_api_keys):
        api_dict = {"ElectricityMaps": "mock_api_key"}
        _tracker = CarbonTracker(epochs=1, api_keys=api_dict)

        mock_set_api_keys.assert_called_once_with(api_dict)

    def test_output_energy(self):
        assert self.tracker is not None
        assert self.mock_logger is not None

        description = "Test description"
        time = 1000
        energy = 50.123
        co2eq = 100.456
        conversions = [(100, "km"), (200, "kg")]

        self.tracker._output_energy(description, time, energy, co2eq, conversions)

        expected_output = (
            "\nTest description\n"
            "\tTime:\t0:16:40\n"
            "\tEnergy:\t50.123000 kWh\n"
            "\tCO2eq:\t100.456000 g"
            "\n\tThis is equivalent to:"
            "\n\t100.000000 km"
            "\n\t200.000000 kg"
        )
        self.mock_logger.output.assert_called_once_with(
            expected_output, verbose_level=1
        )

    def test_output_actual_zero_epochs(self):
        assert self.tracker is not None
        assert self.mock_logger is not None

        self.tracker.epochs_before_pred = 0
        self.tracker.tracker.total_energy_per_epoch = MagicMock(
            return_value=np.array([10, 20, 30])
        )
        self.tracker.tracker.epoch_times = [100, 200, 300]
        self.tracker._co2eq = MagicMock(return_value=150)
        self.tracker.interpretable = True

        self.tracker._output_actual()

        expected_output = (
            "\nActual consumption:\n"
            "\tTime:\t0:10:00\n"
            "\tEnergy:\t60.000000 kWh\n"
            "\tCO2eq:\t150.000000 g"
            "\n\tThis is equivalent to:\n"
            "\t1.395349 km travelled by car"
        )

        self.mock_logger.output.assert_called_once_with(
            expected_output, verbose_level=1
        )

    def test_output_actual_nonzero_epochs(self):
        assert self.tracker is not None
        assert self.mock_logger is not None

        self.tracker.epochs_before_pred = 1
        self.tracker.epoch_counter = 2
        self.tracker.tracker.total_energy_per_epoch = MagicMock(
            return_value=np.array([10, 20, 30])
        )
        self.tracker.tracker.epoch_times = [100, 200, 300]
        self.tracker._co2eq = MagicMock(return_value=150)
        self.tracker.interpretable = True

        self.tracker._output_actual()

        expected_description = "Actual consumption for 2 epoch(s):"

        expected_output = (
            f"\n{expected_description}\n"
            "\tTime:\t0:10:00\n"
            "\tEnergy:\t60.000000 kWh\n"
            "\tCO2eq:\t150.000000 g"
            "\n\tThis is equivalent to:\n"
            "\t1.395349 km travelled by car"
        )

        self.mock_logger.output.assert_called_once_with(
            expected_output, verbose_level=1
        )

    def test_output_pred(self):
        assert self.tracker is not None
        assert self.mock_logger is not None

        predictor = MagicMock()
        predictor.predict_energy = MagicMock(return_value=100)
        predictor.predict_time = MagicMock(return_value=1000)

        self.tracker.epochs = 5
        self.tracker.tracker.total_energy_per_epoch = MagicMock(
            return_value=[10, 20, 30]
        )
        self.tracker.tracker.epoch_times = [100, 200, 300]
        self.tracker._co2eq = MagicMock(return_value=150)
        self.tracker.interpretable = True

        self.tracker._output_pred()

        expected_description = "Predicted consumption for 5 epoch(s):"

        expected_output = (
            f"\n{expected_description}\n"
            "\tTime:\t0:16:40\n"
            "\tEnergy:\t100.000000 kWh\n"
            "\tCO2eq:\t150.000000 g"
            "\n\tThis is equivalent to:\n"
            "\t1.395349 km travelled by car"
        )

        self.mock_logger.output.assert_called_once_with(
            expected_output, verbose_level=1
        )

    def test_co2eq_with_pred_time_dur(self):
        assert self.tracker is not None
        intensity_updater = MagicMock()
        intensity_updater.predict_carbon_intensity = MagicMock(
            return_value=MagicMock(carbon_intensity=0.5)
        )

        energy_usage = 100
        pred_time_dur = 1000

        self.tracker.intensity_updater = intensity_updater

        co2eq = self.tracker._co2eq(energy_usage, pred_time_dur)

        expected_co2eq = 50
        self.assertEqual(co2eq, expected_co2eq)

    def test_co2eq_without_pred_time_dur(self):
        assert self.tracker is not None
        intensity_updater = MagicMock()
        intensity_updater.average_carbon_intensity = MagicMock(
            return_value=MagicMock(carbon_intensity=0.5)
        )

        energy_usage = 100

        self.tracker.intensity_updater = intensity_updater

        co2eq = self.tracker._co2eq(energy_usage)

        expected_co2eq = 50
        self.assertEqual(co2eq, expected_co2eq)

    @patch("sys.exit")
    def test_set_api_keys_with_invalid_name_exits(self, mock_exit):
        assert self.tracker is not None
        self.tracker.set_api_keys({"invalid_name": "test_key"})
        mock_exit.assert_called_once_with(70)

    @mock.patch("carbontracker.tracker.CarbonTracker._get_pids")
    @mock.patch("carbontracker.tracker.loggerutil.Logger")
    @mock.patch("carbontracker.tracker.CarbonTrackerThread")
    @mock.patch("carbontracker.tracker.CarbonIntensityThread")
    def test_exception_handling(
        self, mock_intensity_thread, mock_tracker_thread, mock_logger, mock_get_pids
    ):
        mock_get_pids.side_effect = Exception("Test exception in _get_pids")
        mock_logger.side_effect = Exception("Test exception in Logger initialization")
        mock_tracker_thread.side_effect = Exception(
            "Test exception in CarbonTrackerThread initialization"
        )
        mock_intensity_thread.side_effect = Exception(
            "Test exception in CarbonIntensityThread initialization"
        )

        with self.assertRaises(Exception) as context:
            CarbonTracker(log_dir=None, verbose=False, log_file_prefix="", epochs=1)

        self.assertEqual(
            str(context.exception), "'CarbonTracker' object has no attribute 'logger'"
        )

    # # Instantiating a second instance should not make this instance log twice
    # @mock.patch("carbontracker.tracker.CarbonIntensityThread")
    # def test_multiple_instances(self, mock_intensity_thread):
    #     assert self.mock_logger is not None
    #     assert self.tracker is not None

    #     tracker2 = CarbonTracker(
    #         epochs=5,
    #         epochs_before_pred=1,
    #         monitor_epochs=3,
    #         update_interval=10,
    #         interpretable=True,
    #         stop_and_confirm=True,
    #         ignore_errors=False,
    #         components="all",
    #         devices_by_pid=False,
    #         log_dir=None,
    #         log_file_prefix="",
    #         verbose=1,
    #         decimal_precision=6,
    #     )

    #     predictor = MagicMock()
    #     predictor.predict_energy = MagicMock(return_value=100)
    #     predictor.predict_time = MagicMock(return_value=1000)

    #     self.tracker.epochs = 5
    #     self.tracker.tracker.total_energy_per_epoch = MagicMock(
    #         return_value=[10, 20, 30]
    #     )
    #     self.tracker.tracker.epoch_times = [100, 200, 300]
    #     self.tracker._co2eq = MagicMock(return_value=150)
    #     self.tracker.interpretable = True

    #     self.tracker._output_pred()

    #     expected_description = "Predicted consumption for 5 epoch(s):"

    #     expected_output = (
    #         f"\n{expected_description}\n"
    #         "\tTime:\t0:16:40\n"
    #         "\tEnergy:\t100.000000 kWh\n"
    #         "\tCO2eq:\t150.000000 g"
    #         "\n\tThis is equivalent to:\n"
    #         "\t1.395349 km travelled by car"
    #     )

    #     self.mock_logger.output.assert_called_once_with(
    #         expected_output, verbose_level=1
    #     )


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch, MagicMock
from carbontracker.emissions.intensity.fetchers.electricitymaps import ElectricityMap
from carbontracker import exceptions

class TestElectricityMap(unittest.TestCase):
    def setUp(self):
        self.logger = MagicMock()
        self.electricity_map = ElectricityMap(logger=self.logger)
        self.g_location = MagicMock()
        self.g_location.lng = 0.0
        self.g_location.lat = 0.0
        self.g_location.country = "US"

    def test_set_api_key(self):
        ElectricityMap.set_api_key("test_key")
        self.assertEqual(ElectricityMap._api_key, "test_key")

    def test_suitable(self):
        ElectricityMap.set_api_key("test_key")
        self.assertTrue(self.electricity_map.suitable(self.g_location))

    @patch("requests.get")
    def test_carbon_intensity_by_location_with_lon_lat(self, mock_get):
        mock_response = MagicMock()
        mock_response.ok = True
        mock_response.json.return_value = {"carbonIntensity": 50.0}
        mock_get.return_value = mock_response

        result = self.electricity_map._carbon_intensity_by_location(lon=self.g_location.lng, lat=self.g_location.lat)
        self.assertEqual(result, 50.0)

    @patch("requests.get")
    def test_carbon_intensity_by_location_with_zone(self, mock_get):
        mock_response = MagicMock()
        mock_response.ok = True
        mock_response.json.return_value = {"carbonIntensity": 75.0}
        mock_get.return_value = mock_response

        result = self.electricity_map._carbon_intensity_by_location(zone=self.g_location.country)
        self.assertEqual(result, 75.0)

    @patch("requests.get")
    def test_carbon_intensity_by_location_response_not_ok(self, mock_get):
        mock_response = MagicMock()
        mock_response.ok = False
        mock_response.json.return_value = {"error": "some error"}
        mock_get.return_value = mock_response

        with self.assertRaises(exceptions.CarbonIntensityFetcherError):
            self.electricity_map._carbon_intensity_by_location(lon=self.g_location.lng, lat=self.g_location.lat)

    @patch.object(ElectricityMap, "_carbon_intensity_by_location")
    def test_carbon_intensity(self, mock_carbon_intensity_by_location):
        mock_carbon_intensity_by_location.return_value = 100.0

        result = self.electricity_map.carbon_intensity(self.g_location)

        self.assertEqual(result.carbon_intensity, 100.0)

    @patch.object(ElectricityMap, "_carbon_intensity_by_location")
    def test_carbon_intensity_with_exception(self, mock_carbon_intensity_by_location):
        mock_carbon_intensity_by_location.side_effect = [Exception(), 25.0]

        result = self.electricity_map.carbon_intensity(self.g_location)

        mock_carbon_intensity_by_location.assert_called_with(zone=self.g_location.country)
        self.assertEqual(result.carbon_intensity, 25.0)


if __name__ == "__main__":
    unittest.main()